#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
# \package test nexdatas
# \file XMLConfigurator_test.py
# unittests for field Tags running Tango Server
#
import unittest
import sys
import shutil
import os


try:
    import NXSCreatePoolDSFS_test
except Exception:
    from . import NXSCreatePoolDSFS_test


if sys.version_info > (3,):
    unicode = str
    long = int


# test fixture
class NXSCreatePoolDSFS2Test(
        NXSCreatePoolDSFS_test.NXSCreatePoolDSFSTest):

    # constructor
    # \param methodName name of the test method
    def __init__(self, methodName):
        NXSCreatePoolDSFS_test.NXSCreatePoolDSFSTest.__init__(
            self, methodName)

        self.directory = "my_test_nxs"
        self._dircreated = False
        self.flags = " -d %s" % self.directory

    # test starter
    # \brief Common set up
    def setUp(self):
        NXSCreatePoolDSFS_test.NXSCreatePoolDSFSTest.setUp(self)
        if not os.path.isdir(self.directory):
            os.makedirs(self.directory)
            self._dircreated = True

    # test closer
    # \brief Common tear down
    def tearDown(self):
        NXSCreatePoolDSFS_test.NXSCreatePoolDSFSTest.tearDown(self)
        if self._dircreated:
            shutil.rmtree(self.directory)
            self._dircreated = False

    def test_poolds_pdevice_motors_channels_file_prefix(self):
        """ test nxsccreate poolds file system
        """
        fun = sys._getframe().f_code.co_name
        print("Run: %s.%s() " % (self.__class__.__name__, fun))

        self._tsv.dp.MotorList = [
            '{'
            '"name": "tst_exp_mot01", '
            '"source" : "tango://haso.desy.de:10000/'
            'motor/oms/10/Position",'
            '"type": "Motor" '
            '}',
            '{'
            '"name": "tst_exp_mot02", '
            '"source" : "tango://haso.desy.de:10000/'
            'pmotor/oms/20/Position",'
            '"type": "PseudoMotor" '
            '}'

        ]
        self._tsv.dp.ExpChannelList = [
            '{'
            '"name": "tst_exp_c01", '
            '"source" : "tango://haso.desy.de:10000/'
            'p00/counter/10/Counts",'
            '"type": "CTExpChannel" '
            '}',
            '{'
            '"name": "tst_exp_adc01", '
            '"source" : "tango://haso.desy.de:10000/'
            'p00/adc/10/Value",'
            '"type": "ZeroDExpChannel" '
            '}',
            '{'
            '"name": "tst_exp_mca01", '
            '"source" : "tango://haso.desy.de:10000/'
            'p00/mca/10/Value",'
            '"type": "OneDExpChannel" '
            '}',
            '{'
            '"name": "tst_exp_det01", '
            '"source" : "tango://haso.desy.de:10000/'
            'p00/det/10/Data",'
            '"type": "TwoDExpChannel" '
            '}',
            '{'
            '"name": "tst_exp_pc01", '
            '"source" : "tango://haso.desy.de:10000/'
            'p00/pcounter/10/Counts",'
            '"type": "PseudoCounter" '
            '}',
        ]
        args = [
            [
                ('nxscreate poolds -x test_ -p %s %s '
                 % (self._tsv.dp.name(), self.flags)).split(),
                ('nxscreate poolds --file-prefix test_ --pool %s %s '
                 % (self._tsv.dp.name(), self.flags)).split()
            ],
            [
                'test_tst_exp_mot01',
                'test_tst_exp_mot02',
                'test_tst_exp_c01',
                'test_tst_exp_adc01',
                'test_tst_exp_mca01',
                'test_tst_exp_det01',
                'test_tst_exp_pc01',
            ],
            [
                '<?xml version=\'1.0\' encoding=\'utf8\'?>\n'
                '<definition>\n'
                '  <datasource type="TANGO" name="tst_exp_mot01">\n'
                '    <device name="motor/oms/10" '
                'member="attribute" hostname="haso.desy.de" '
                'port="10000" group="__CLIENT__"/>\n'
                '    <record name="Position"/>\n'
                '  </datasource>\n'
                '</definition>\n',
                '<?xml version=\'1.0\' encoding=\'utf8\'?>\n'
                '<definition>\n'
                '  <datasource type="TANGO" name="tst_exp_mot02">\n'
                '    <device name="pmotor/oms/20" '
                'member="attribute" hostname="haso.desy.de" '
                'port="10000" group="__CLIENT__"/>\n'
                '    <record name="Position"/>\n'
                '  </datasource>\n'
                '</definition>\n',

                '<?xml version=\'1.0\' encoding=\'utf8\'?>\n'
                '<definition>\n'
                '  <datasource type="TANGO" name="tst_exp_c01">\n'
                '    <device name="p00/counter/10" '
                'member="attribute" hostname="haso.desy.de" '
                'port="10000" group="__CLIENT__"/>\n'
                '    <record name="Counts"/>\n'
                '  </datasource>\n'
                '</definition>\n',
                '<?xml version=\'1.0\' encoding=\'utf8\'?>\n'
                '<definition>\n'
                '  <datasource type="TANGO" name="tst_exp_adc01">\n'
                '    <device name="p00/adc/10" '
                'member="attribute" hostname="haso.desy.de" '
                'port="10000" group="__CLIENT__"/>\n'
                '    <record name="Value"/>\n'
                '  </datasource>\n'
                '</definition>\n',
                '<?xml version=\'1.0\' encoding=\'utf8\'?>\n'
                '<definition>\n'
                '  <datasource type="TANGO" name="tst_exp_mca01">\n'
                '    <device name="p00/mca/10" '
                'member="attribute" hostname="haso.desy.de" '
                'port="10000" group="__CLIENT__"/>\n'
                '    <record name="Value"/>\n'
                '  </datasource>\n'
                '</definition>\n',
                '<?xml version=\'1.0\' encoding=\'utf8\'?>\n'
                '<definition>\n'
                '  <datasource type="TANGO" name="tst_exp_det01">\n'
                '    <device name="p00/det/10" '
                'member="attribute" hostname="haso.desy.de" '
                'port="10000" group="__CLIENT__"/>\n'
                '    <record name="Data"/>\n'
                '  </datasource>\n'
                '</definition>\n',
                '<?xml version=\'1.0\' encoding=\'utf8\'?>\n'
                '<definition>\n'
                '  <datasource type="TANGO" name="tst_exp_pc01">\n'
                '    <device name="p00/pcounter/10" '
                'member="attribute" hostname="haso.desy.de" '
                'port="10000" group="__CLIENT__"/>\n'
                '    <record name="Counts"/>\n'
                '  </datasource>\n'
                '</definition>\n',
            ],
        ]

        totest = []
        try:
            for cmd in args[0]:
                skip = False
                for ds in args[1]:
                    if self.dsexists(ds):
                        skip = True
                if not skip:
                    for ds in args[1]:
                        totest.append(ds)

                    vl, er = self.runtest(cmd)
                    if er:
                        self.assertTrue(er.startswith("Info: "))
                    else:
                        self.assertEqual('', er)
                    self.assertTrue(vl)

                    for i, ds in enumerate(args[1]):
                        xml = self.getds(ds)
                        self.assertEqual(
                            args[2][i], xml)

                    for ds in args[1]:
                        self.deleteds(ds)
        finally:
            for ds in totest:
                if self.dsexists(ds):
                    self.deleteds(ds)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

"""  pyeval helper functions for xspress """

import os
try:
    import tango
except Exception:
    import PyTango as tango


def triggermode_cb(commonblock, name, triggermode,
                   nbframes, hostname, device,
                   filename, entryname, insname,
                   filedir, fileprefix, framesperfile,
                   maskdatatowrite, mcalength, savedata,
                   acq_modes=""):
    """ code for triggermode_cb  datasource

    :param commonblock: commonblock of nxswriter
    :type commonblock: :obj:`dict`<:obj:`str`, `any`>
    :param name: component name
    :type name: :obj:`str`
    :param triggermode:  trigger mode
    :type triggermode: :obj:`int`
    :param nbframes: a number of images
    :type nbframes: :obj:`int`
    :param hostname: tango host name
    :type hostname: :obj:`str`
    :param device: tango device name
    :type device: :obj:`str`
    :param filename: file name
    :type filename: :obj:`str`
    :param entryname: entry name
    :type entryname: :obj:`str`
    :param insname: instrument name
    :type insname: :obj:`str`
    :param filedir: file directory
    :type filedir: :obj:`str`
    :param fileprefix: file prefix
    :type fileprefix: :obj:`str`
    :param framesperfile: frame number per file
    :type framesperfile: :obj:`int`
    :param maskdatatowrite: mask data to write
    :type maskdatatowrite: :obj:`int`
    :param mcalength: mcalength
    :type mcalength: :obj:`int`
    :param savedata: savedata flag
    :type savedata: :obj:`bool`
    :param acq_modes: acquisition modes
    :type acq_modes: :obj:`str`
    :returns: triggermode
    :rtype: :obj:`int`
    """

    if not savedata:
        return triggermode
    amodes = acq_modes.split(",")
    xp = tango.DeviceProxy('%s/%s' % (hostname, device))
    #  print("device", device, hostname)
    try:
        nbchannels = int(xp.get_property("NbChannels")["NbChannels"][0])
        # print("GET", int(xp.get_property("NbChannels")["NbChannels"][0]))
    except Exception as e:
        print("EXCE", str(e))
        nbchannels = 0

    if not nbchannels:
        return triggermode

    if "__root__" in commonblock.keys():
        root = commonblock["__root__"]
        if root.h5object.__class__.__name__ == "File":
            import nxstools.h5pywriter as nxw
        else:
            import nxstools.h5cppwriter as nxw
    else:
        raise Exception("Writer cannot be found")

    path = ""
    sfname = []
    if not filename:
        if root._tparent is not None:
            filename = root._tparent.filename
    basepath = ""
    if filename:
        sfname = (filename).split("/")
        path = sfname[-1].split(".")[0] + "/"
        basepath = "/".join(os.path.abspath(filename).split("/")[:-1])
    if filedir and filedir.startswith(basepath):
        path = filedir[len(basepath) + 1:]
    else:
        path += '%s' % (name)
    path += '/%s_' % (fileprefix)

    en = root.open(entryname)
    dt = en.open("data")
    ins = en.open(insname)

    nbfiles = (nbframes + framesperfile - 1) // framesperfile
    shape = [nbframes, mcalength]
    dtype = "int32"

    for nch in range(nbchannels):
        masked = maskdatatowrite & (1 << nch)
        # print("NCHAN", nch, masked)
        if masked:
            continue
        chname = "%s_channel%02i" % (name, nch)
        detc = ins.create_group(chname, "NXdetector")
        colc = detc.create_group("collection", "NXcollection")

        if "VDS" in amodes and "data" not in detc.names():
            vfl = nxw.virtual_field_layout(shape, dtype)

        for nbf in range(nbfiles):
            nxw.link(
                "%s%05i.nxs://entry/instrument/xspress3/channel%02i/histogram"
                % (path, nbf, nch),
                colc, "data_%05i" % (nbf))
            nxw.link("/%s/%s/%s/collection/%s" %
                     (entryname, insname, chname, "data_%05i"
                      % (nbf)), dt,
                     "%s_channel%02i_%05i" % (name, nch, nbf))

            if "VDS" in amodes and "data" not in detc.names():
                off = framesperfile * nbf
                if nbf + 1 == nbfiles:
                    nb = nbframes - off
                else:
                    nb = framesperfile
                ef = nxw.target_field_view(
                    "%s%05i.nxs" % (path, nbf),
                    "/entry/instrument/xspress3/channel%02i/histogram" % (nch),
                    [nb, shape[1]], dtype)
                vfl.add(
                    (slice(off, off + nb), slice(0, shape[1])),
                    ef, (slice(None), slice(None)))
        if "VDS" in amodes and "data" not in detc.names():
            detc.create_virtual_field("data", vfl)
        elif nbfiles == 1:
            nxw.link("/%s/%s/%s/collection/%s" %
                     (entryname, insname, chname, "data_%05i"
                      % (0)), detc, "data")
        if chname not in dt.names() and "data" in detc.names():
            nxw.link("/%s/%s/%s/data" % (entryname, insname, chname),
                     dt, chname)

    return triggermode



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

"""  pyeval helper functions for limaccd """

import os


def postrun(commonblock,
            saving_next_number,
            saving_directory,
            saving_suffix,
            acq_nb_frames,
            saving_index_format,
            saving_prefix,
            saving_next_number_str,
            name=None,
            saving_format=None,
            saving_frame_per_file=None,
            image_height=None,
            image_width=None,
            image_type=None,
            acq_trigger_mode=None,
            acq_mode='SINGLE',
            filename=None,
            entryname=None,
            insname=None,
            acq_modes="",
            field_path="/entry_0000/measurement/data"
            ):
    """ code for postrun datasource

    :param commonblock: commonblock of nxswriter
    :type commonblock: :obj:`dict`<:obj:`str`, `any`>
    :param saving_next_number: saving next number
    :type saving_next_number: :obj:`int`
    :param saving_directory: saving directory
    :type saving_directory: :obj:`str`
    :param saving_suffix: saving suffix
    :type saving_suffix: :obj:`str`
    :param acq_nb_frames: number of frames acquired
    :type acq_nb_frames: :obj:`str`
    :param saving_index_format: saving index format
    :type saving_index_format: :obj:`str`
    :param saving_prefix: saving prefix
    :type saving_prefix: :obj:`str`
    :param saving_next_number_str: datasource string name
    :type saving_next_number_str: :obj:`str`
    :param name: component name
    :type name: :obj:`str`
    :param saving_format: saving format
    :type saving_format: :obj:`str`
    :param saving_frame_per_file: saving frame per file
    :type saving_frame_per_file: :obj:`str`
    :param image_height: image height
    :type image_height: :obj:`int`
    :param image_width: image width
    :type image_width: :obj:`int`
    :param image_type: image type
    :type image_type: :obj:`int`
    :param acq_mode: acquisition mode
    :type acq_mode: :obj:`str`
    :param acq_trigger_mode: acquisition trigger mode
    :type acq_trigger_mode: :obj:`str`
    :param filename: file name
    :type filename: :obj:`str`
    :param entryname: entry name
    :type entryname: :obj:`str`
    :param insname: instrument name
    :type insname: :obj:`str`
    :param acq_modes: acquisition modes
    :type acq_modes: :obj:`str`
    :param field_path: nexus field path
    :type field_path: :obj:`str`
    :returns: name of saved file
    :rtype: :obj:`str`
    """
    filedir = (saving_directory).replace("\\", "/")
    amodes = acq_modes.split(",")
    if len(filedir) > 1 and filedir[1] == ":":
        filedir = "/data" + filedir[2:]
    if filedir and filedir[-1] == "/":
        filedir = filedir[:-1]
    if acq_mode == "SINGLE" and \
            acq_trigger_mode in ["INTERNAL_TRIGGER_MULTI",
                                 "EXTERNAL_TRIGGER_MULTI"] and \
            saving_format == "HDF5":
        filelastnumber = saving_next_number - 1
        nbfiles = (acq_nb_frames + saving_frame_per_file - 1) \
            // saving_frame_per_file
        filestartnum = filelastnumber - nbfiles + 1

        if "__root__" in commonblock.keys():
            root = commonblock["__root__"]
            if root.h5object.__class__.__name__ == "File":
                import nxstools.h5pywriter as nxw
            else:
                import nxstools.h5cppwriter as nxw
        else:
            raise Exception("Writer cannot be found")

        path = ""
        sfname = []
        if not filename:
            if root._tparent is not None:
                filename = root._tparent.filename
        basepath = ""
        if filename:
            sfname = (filename).split("/")
            path = sfname[-1].split(".")[0] + "/"
            basepath = "/".join(os.path.abspath(filename).split("/")[:-1])
        if filedir and filedir.startswith(basepath):
            path = filedir[len(basepath) + 1:]
        else:
            path += '%s' % (name)
        path += '/%s' % (saving_prefix)

        en = root.open(entryname)
        dt = en.open("data")
        ins = en.open(insname)
        det = ins.open(name)
        try:
            col = det.open("collection")
        except Exception:
            col = det.create_group("collection", "NXcollection")

        shape = [image_height, image_width]
        if "VDS" in amodes and "data" not in det.names():
            l2nt = {
                "bpp8": "uint8", "bpp8s": "int8",
                "bpp16": "uint16", "bpp16s": "int16",
                "bpp32": "uint32", "bpp32s": "int32",
                "bpp32f": "float",
            }
            dtype = l2nt.get(image_type.lower(), "int32")
            vfl = nxw.virtual_field_layout(
                [acq_nb_frames, shape[0], shape[1]], dtype)

        fnamepattern = "%s%s%s" % (path, saving_index_format, saving_suffix)

        for nbf in range(filestartnum, filelastnumber + 1):
            rnbf = nbf - filestartnum
            try:
                detfn = fnamepattern % nbf
            except Exception:
                detfn = fnamepattern

            nxw.link("%s:/%s" % (detfn, field_path),
                     col, "data_%05i" % (nbf))
            nxw.link("/%s/%s/%s/collection/%s" %
                     (entryname, insname, name, "data_%05i" % (nbf)), dt,
                     "%s_%05i" % (name, nbf))

            if "VDS" in amodes and "data" not in det.names():
                off = saving_frame_per_file * rnbf
                if rnbf + 1 == nbfiles:
                    nb = acq_nb_frames - off
                else:
                    nb = saving_frame_per_file
                ef = nxw.target_field_view(
                    detfn, field_path,
                    [nb, shape[0], shape[1]], dtype)
                vfl.add(
                    (slice(off, off + nb),
                     slice(0, shape[0]),
                     slice(0, shape[1])),
                    ef, (slice(None), slice(None), slice(None)))
        if "VDS" in amodes and "data" not in det.names():
            det.create_virtual_field("data", vfl)
        elif nbfiles == 1:
            nxw.link("/%s/%s/%s/collection/%s" %
                     (entryname, insname, name, "data_%05i"
                      % (filestartnum)), det, "data")
        if name not in dt.names() and "data" in det.names():
            nxw.link("/%s/%s/%s/data" % (entryname, insname, name),
                     dt, name)
        result = ""
    else:
        filestartnum = commonblock[saving_next_number_str] - 1
        result = "" + filedir + "/" + saving_prefix + saving_index_format
        result += saving_suffix + ":"
        filelastnumber = saving_next_number - 1
        if acq_mode == "SINGLE" and \
                acq_trigger_mode in ["INTERNAL_TRIGGER_MULTI",
                                     "EXTERNAL_TRIGGER_MULTI"]:
            nbfiles = (acq_nb_frames + saving_frame_per_file - 1) \
                // saving_frame_per_file
            filestartnum = filelastnumber - nbfiles + 1
        if "__root__" in commonblock.keys():
            root = commonblock["__root__"]
            if hasattr(root, "currentfileid") and \
                    hasattr(root, "stepsperfile"):
                spf = root.stepsperfile
                cfid = root.currentfileid
                if spf > 0 and cfid > 0:
                    filelastnumber = min(
                        filestartnum + cfid * acq_nb_frames * spf - 1,
                        filelastnumber)
                    filestartnum = filestartnum + (cfid - 1) \
                        * acq_nb_frames * spf
        result += str(filestartnum) + ":" + str(filelastnumber)
    return result



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

"""  pyeval helper functions for cobold """

from nxstools import filewriter


def time_of_flight(commonblock, binsize, entryname, histogram):
    """ code for time_of_flight  datasource

    :param commonblock: commonblock of nxswriter
    :type commonblock: :obj:`dict`<:obj:`str`, `any`>
    :param detector: bin size
    :type detector: :obj:`float`
    :param entryname:  entry group name
    :type entryname: :obj:`str`
    :param histogram:  histogram field name
    :type histogram: :obj:`str`
    :returns: time of flight axis data
    :rtype: :obj:`list` <:obj:`float`>
    """

    result = []
    fld = None
    names = []
    try:
        binsize = float(filewriter.first(binsize))

        root = commonblock["__root__"]
        nxentry = root.open(entryname)
        nxdata = nxentry.open("data")
        writer = root.parent.writer
        links = writer.get_links(nxdata)
        names = list(sorted([ch.name for ch in links]))

        if histogram in names:
            fld = nxdata.open(histogram)
            if isinstance(fld, filewriter.FTField) and \
               len(fld.shape) == 2 and fld.shape[1] > 0:
                result = fld.shape
                result = [it * binsize for it in range(fld.shape[1])]
    except Exception:
        result = []
    return result



#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

""" Set Up NeXus Tango Servers"""

import socket
import os
import sys
import time
import json
import argparse
import subprocess

try:
    import tango
except Exception:
    import PyTango as tango

from nxstools.nxsargparser import (Runner, NXSArgParser, ErrorException)

#: (:obj:`str`) host name
_hostname = socket.gethostname()


#: (:obj:`dict` <:obj:`dict` <:obj:`str` , :obj:`str` > > )
#:    all SardanaHosts and DataBaseHosts should be known
knownHosts = {
    'hasdelay': {'beamline': 'delay', 'masterhost': 'hasdelay',
                 'user': 'delayusr', 'dbname': 'nxsconfig'},
    'hasdelay2': {'beamline': 'delsauto', 'masterhost': 'hasdelay2',
                  'user': 'delayusr', 'dbname': 'nxsconfig'},
    'hasfpgm1': {'beamline': 'flash', 'masterhost': 'hasfpgm1',
                 'user': 'vuvfuser', 'dbname': 'nxsconfig'},
    'hasfvls': {'beamline': 'flash', 'masterhost': 'hasfvls',
                'user': 'vuvfuser', 'dbname': 'nxsconfig'},
    'hasfmirr': {'beamline': 'flash', 'masterhost': 'hasfmirr',
                 'user': 'musixusr', 'dbname': 'nxsconfig'},
    'haskmusixtng': {'beamline': 'flash', 'masterhost': 'haskmusixtng',
                     'user': 'vuvfuser', 'dbname': 'nxsconfig'},
    'hasmfmc': {'beamline': 'fmc', 'masterhost': 'hasmfmc',
                'user': 'delayusr', 'dbname': 'nxsconfig'},
    'hasmlqj': {'beamline': 'lqj', 'masterhost': 'hasmlqj',
                'user': 'lqjuser', 'dbname': 'nxsconfig'},
    # ? 'haso052metro': {'beamline': 'metro', 'masterhost': 'haso052metro',
    # ?            'user': '???', 'dbname': 'nxsconfig'},
    'hasodlsauto': {'beamline': 'dlsauto', 'masterhost': 'hasodlsauto',
                    'user': 'dlsuser', 'dbname': 'nxsconfig'},
    'haso111n': {'beamline': 'p09', 'masterhost': 'haso111n',
                 'user': 'tnunez', 'dbname': 'nxsconfig'},
    'haso111tb': {'beamline': 'p09', 'masterhost': 'haso111tb',
                  'user': 'tnunez', 'dbname': 'nxsconfig'},
    'haso113b': {'beamline': 'p09', 'masterhost': 'haso113b',
                 'user': 'blume', 'dbname': 'nxsconfig'},
    'haso111o': {'beamline': 'p09'},  # Benjamin Obuna
    'haso113u': {'beamline': 'p09', 'masterhost': 'haso113u',
                 'user': 'kracht', 'dbname': 'nxsconfig'},
    'hastodt': {'beamline': 'p09', 'masterhost': 'hastodt',
                'user': 'kracht', 'dbname': 'nxsconfig'},
    'haso204n': {'beamline': 'p23', 'masterhost': 'haso204n',
                 'user': 'p23user', 'dbname': 'nxsconfig'},
    'haso228jk': {'beamline': 'p09', 'masterhost': 'haso228jk',
                  'user': 'jkotan', 'dbname': 'nxsconfig'},
    'haso213p': {'beamline': 'p22', 'masterhost': 'haso213p',
                 'user': 'spiec', 'dbname': 'nxsconfig'},
    'has6117b': {'beamline': 'p02', 'masterhost': 'has6117b',
                 'user': 'p02user', 'dbname': 'nxsconfig'},
    'haspecsicl4': {'beamline': 'p02', 'masterhost': 'haspecsicl4',
                    'user': 'lacluser', 'dbname': 'nxsconfig'},
    'haspllabcl1': {'beamline': 'llab', 'masterhost': 'haspllabcl1',
                    'user': 'lacluser', 'dbname': 'nxsconfig'},
    # ?'hasp144p': {'beamline': 'p11', 'masterhost': 'hasp144p',
    # ?               'user': '???', 'dbname': 'nxsconfig'},
    'haspp01eh1': {'beamline': 'p01', 'masterhost': 'haspp01eh1',
                   'user': 'p01user', 'dbname': 'nxsconfig'},
    'haspp01eh2': {'beamline': 'p01', 'masterhost': 'haspp01eh2',
                   'user': 'p01user', 'dbname': 'nxsconfig'},
    'haspp01eh3': {'beamline': 'p01', 'masterhost': 'haspp01eh3',
                   'user': 'p01user', 'dbname': 'nxsconfig'},
    'haspp02ch1a': {'beamline': 'p02', 'masterhost': 'haspp02ch1a',
                    'user': 'p02user', 'dbname': 'nxsconfig'},
    'haspp021ch1a': {'beamline': 'p021', 'masterhost': 'haspp021ch1a',
                     'user': 'p021user', 'dbname': 'nxsconfig'},
    'haspp02ch1': {'beamline': 'p02', 'masterhost': 'haspp02ch1',
                   'user': 'p02user', 'dbname': 'nxsconfig'},
    'haspp021ch1': {'beamline': 'p02', 'masterhost': 'haspp021ch1',
                    'user': 'p021user', 'dbname': 'nxsconfig'},
    'haspp02ch2': {'beamline': 'p02', 'masterhost': 'haspp02ch2',
                   'user': 'p02user', 'dbname': 'nxsconfig'},
    'haspp02oh1': {'beamline': 'p02', 'masterhost': 'haspp02oh1',
                   'user': 'p02user', 'dbname': 'nxsconfig'},
    'haspp02lakl': {'beamline': 'p02', 'masterhost': 'haspp02lakl',
                    'user': 'p02user', 'dbname': 'nxsconfig'},
    'haspp022ch': {'beamline': 'p022', 'masterhost': 'haspp022ch',
                   'user': 'p022user', 'dbname': 'nxsconfig'},
    'haspp022chms': {'beamline': 'p022', 'masterhost': 'haspp022chms',
                     'user': 'p022user', 'dbname': 'nxsconfig'},
    'haso224w': {'beamline': 'p02', 'masterhost': 'haso224w',
                 'user': 'p021user', 'dbname': 'nxsconfig'},
    'haso232s': {'beamline': 'p02', 'masterhost': 'haspp232s',
                 'user': 'p02user', 'dbname': 'nxsconfig'},
    'haspp021jenkins': {'beamline': 'p021', 'masterhost': 'haspp021jenkins',
                        'user': 'p021user', 'dbname': 'nxsconfig'},
    'haspp03': {'beamline': 'p03', 'masterhost': 'haspp03',
                'user': 'p03user', 'dbname': 'nxsconfig'},
    'haspp03nano': {'beamline': 'p03nano', 'masterhost': 'haspp03nano',
                    'user': 'p03nano', 'dbname': 'nxsconfig'},
    'haspp04exp1': {'beamline': 'p04', 'masterhost': 'haspp04exp1',
                    'user': 'p04user', 'dbname': 'nxsconfig'},
    'haspp04exp2': {'beamline': 'p04', 'masterhost': 'haspp04exp2',
                    'user': 'p04user', 'dbname': 'nxsconfig'},
    'haspp06ctrl': {'beamline': 'p06', 'masterhost': 'haspp06ctrl',
                    'user': 'p06user', 'dbname': 'nxsconfig'},
    'haspp06mc01': {'beamline': 'p06', 'masterhost': 'haspp06mc01',
                    'user': 'p06user', 'dbname': 'nxsconfig'},
    'haspp06nc1': {'beamline': 'p06', 'masterhost': 'haspp06nc1',
                   'user': 'p06user', 'dbname': 'nxsconfig'},
    'hasp029rack': {'beamline': 'p06', 'masterhost': 'hasp029rack',
                    'user': 'p06user', 'dbname': 'nxsconfig'},
    'hasp068xlab': {'beamline': 'p06', 'masterhost': 'hasp068xlab',
                    'user': 'p06user', 'dbname': 'nxsconfig'},
    'haspp07eh2': {'beamline': 'p07', 'masterhost': 'haspp07eh2',
                   'user': 'p07user', 'dbname': 'nxsconfig'},
    'haspp08': {'beamline': 'p08', 'masterhost': 'haspp08',
                'user': 'p08user', 'dbname': 'nxsconfig'},
    'haspp08lisa2': {'beamline': 'p08', 'masterhost': 'haspp08lisa2',
                     'user': 'p08user', 'dbname': 'nxsconfig'},
    'haspp09': {'beamline': 'p09', 'masterhost': 'haspp09',
                'user': 'p09user', 'dbname': 'nxsconfig'},
    'haspp09dif': {'beamline': 'p09', 'masterhost': 'haspp09dif',
                   'user': 'p09user', 'dbname': 'nxsconfig'},
    'haspp09mag': {'beamline': 'p09', 'masterhost': 'haspp09mag',
                   'user': 'p09user', 'dbname': 'nxsconfig'},
    'haspp09haxps': {'beamline': 'p09', 'masterhost': 'haspp09maxps',
                     'user': 'p09haxps', 'dbname': 'nxsconfig'},
    'haspp10e1': {'beamline': 'p10', 'masterhost': 'haspp10e1',
                  'user': 'p10user', 'dbname': 'nxsconfig'},
    'haspp10e2': {'beamline': 'p10', 'masterhost': 'haspp10e2',
                  'user': 'p10user', 'dbname': 'nxsconfig'},
    'haspp10lcx': {'beamline': 'p10', 'masterhost': 'haspp10lcx',
                   'user': 'p10user', 'dbname': 'nxsconfig'},
    'haspp10lab': {'beamline': 'p10', 'masterhost': 'haspp10lab',
                   'user': 'p10user', 'dbname': 'nxsconfig'},
    'haspp11oh': {'beamline': 'p11', 'masterhost': 'haspp11oh',
                  'user': 'p11user', 'dbname': 'nxsconfig'},
    'haspp11sardana': {'beamline': 'p11',
                       'masterhost': 'haspp11sardana',
                       'user': 'p11user', 'dbname': 'nxsconfig'},
    'haspp11user02': {'beamline': 'p11', 'masterhost': 'haspp11user02',
                      'user': 'p11user', 'dbname': 'nxsconfig'},
    'hasep212lab': {'beamline': 'p21', 'masterhost': 'hasep212lab',
                    'user': 'p212user', 'dbname': 'nxsconfig'},
    'haspp21eh2': {'beamline': 'p21', 'masterhost': 'hasep21eh2',
                   'user': 'p21user', 'dbname': 'nxsconfig'},
    'haspp21eh3': {'beamline': 'p21', 'masterhost': 'hasep21eh3',
                   'user': 'p21user', 'dbname': 'nxsconfig'},
    'haspp212oh': {'beamline': 'p21', 'masterhost': 'hasep212oh',
                   'user': 'p21user', 'dbname': 'nxsconfig'},
    'haspp21lab': {'beamline': 'p21', 'masterhost': 'haspp21lab',
                   'user': 'p21user', 'dbname': 'nxsconfig'},
    'hasep211eh': {'beamline': 'p211', 'masterhost': 'hasep211eh',
                   'user': 'p211user', 'dbname': 'nxsconfig'},
    'hasep212oh': {'beamline': 'p21', 'masterhost': 'hasep212oh',
                   'user': 'p212user', 'dbname': 'nxsconfig'},
    'hasep22oh': {'beamline': 'p22', 'masterhost': 'hasep22oh',
                  'user': 'p22user', 'dbname': 'nxsconfig'},
    'hasep22ch1': {'beamline': 'p22', 'masterhost': 'hasep22ch1',
                   'user': 'p22user', 'dbname': 'nxsconfig'},
    'hasep22ch2': {'beamline': 'p22', 'masterhost': 'hasep22ch2',
                   'user': 'p22user', 'dbname': 'nxsconfig'},
    'hasep23dev': {'beamline': 'p23', 'masterhost': 'hasep23dev',
                   'user': 'p23user', 'dbname': 'nxsconfig'},
    'hasep23eh': {'beamline': 'p23', 'masterhost': 'hasep23eh',
                  'user': 'p23user', 'dbname': 'nxsconfig'},
    # 'hasep23oh': {'beamline': 'p23', 'masterhost': 'hasep23oh',
    #               'user': 'p23user', 'dbname': 'nxsconfig'},
    'hasep23ch': {'beamline': 'p23', 'masterhost': 'hasep23ch',
                  'user': 'p23user', 'dbname': 'nxsconfig'},
    #  ? 'hase027n': {'beamline': 'p24'},
    'hasep24': {'beamline': 'p24', 'masterhost': 'hasep24',
                'user': 'p24user', 'dbname': 'nxsconfig'},
    'hasep24eh1': {'beamline': 'p24', 'masterhost': 'hasep24eh1',
                   'user': 'p24user', 'dbname': 'nxsconfig'},
    'haso107klx': {'beamline': 'p09', 'masterhost': 'haso107klx',
                   'user': 'kracht', 'dbname': 'nxsconfig'},
    'haso107d1': {'beamline': 'p09', 'masterhost': 'haso107d1',
                  'user': 'kracht', 'dbname': 'nxsconfig'},
    'haso107tk': {'beamline': 'p09', 'masterhost': 'haso107tk',
                  'user': 'kracht', 'dbname': 'nxsconfig'},
    # ? 'vbtkdesy': {'beamline': 'p09'},
    # ? 'vbtkhome': {'beamline': 'p09'},
    'hascmexp': {'beamline': 'cmexp', 'masterhost': 'hascmexp',
                 'user': 'cmexp', 'dbname': 'nxsconfig'},
    'hasnp61ch1': {'beamline': 'p61', 'masterhost': 'hasnp61ch1',
                   'user': 'p61user', 'dbname': 'nxsconfig'},
    'hasnp64': {'beamline': 'p64', 'masterhost': 'hasnp64',
                'user': 'p64user', 'dbname': 'nxsconfig'},
    'hasnp64oh': {'beamline': 'p64', 'masterhost': 'hasnp64oh',
                  'user': 'p64user', 'dbname': 'nxsconfig'},
    'hasnp65': {'beamline': 'p65', 'masterhost': 'hasnp65',
                'user': 'p65user', 'dbname': 'nxsconfig'},
    'hasnp66': {'beamline': 'p66', 'masterhost': 'hasnp66',
                'user': 'p66user', 'dbname': 'nxsconfig'},
    'hzgpp07eh1': {'beamline': 'p07', 'masterhost': 'hzgpp07eh1',
                   'user': 'p07user', 'dbname': 'nxsconfig'},
    'hzgpp07eh3': {'beamline': 'p07', 'masterhost': 'hzgpp07eh3',
                   'user': 'p07user', 'dbname': 'nxsconfig'},
    'hzgpp07eh4': {'beamline': 'p07', 'masterhost': 'hzgpp07eh4',
                   'user': 'p07user', 'dbname': 'nxsconfig'},
    'cfeld-pcx27083': {'beamline': 'cfeld', 'masterhost': 'cfeld-pcx27083',
                       'user': 'ritzkowf', 'dbname': 'nxsconfig'},
    'hasmrixs': {'beamline': 'rix', 'masterhost': 'hasmrixs',
                 'user': 'rixuser', 'dbname': 'nxsconfig'},
}


class SetUp(object):

    """ setup NXSDataWriter, NXSConfigServer and NXSRecSelector Tango servers
    """

    def __init__(self):
        """ constructor
        """
        try:
            #: (:class:`tango.Database`) tango database server
            self.db = tango.Database()
        except Exception:
            print("Can't connect to tango database on %s" %
                  os.getenv('TANGO_HOST'))
            sys.exit(255)

        #: (:obj:`str`) NeXus writer device name
        self.writer_name = None
        #: (:obj:`str`) NeXus config server device name
        self.cserver_name = None

    def changeRecorderPath(self, path, instance=None):
        """ adds a new recorder path

        :param path: new recorder path
        :type path: :obj:`str`
        :param instance: MacroServer instance name
        :type instance: :obj:`str`
        :returns: True if record path was added
        :rtype: :obj:`bool`
        """
        res = False
        if not os.path.isdir(path):
            return res
        instance = instance or "*"

        mss = self.db.get_server_list("MacroServer/%s" % instance).value_string
        for ms in mss:
            devserv = self.db.get_device_class_list(ms).value_string
            dev = devserv[0::2]
            serv = devserv[1::2]
            for idx, ser in enumerate(serv):
                if ser == 'MacroServer':
                    if dev[idx]:
                        recorderpaths = self.db.get_device_property(
                            dev[idx], "RecorderPath")["RecorderPath"]
                        if recorderpaths:
                            recorderpaths = [p for p in recorderpaths if p]
                        else:
                            recorderpaths = []
                        if path not in recorderpaths:
                            recorderpaths.append(path)
                            self.db.put_device_property(
                                dev[idx],
                                {"RecorderPath": recorderpaths})
                            res = True
        # time.sleep(0.2)
        return res

    def changePropertyName(self, server, oldname, newname, sclass=None):
        """ changes property name

        :param server: server name
        :type server: :obj:`str`
        :param oldname: old property name
        :type oldname: :obj:`str`
        :param newname: new property name
        :type newname: :obj:`str`
        :param sclass: server class name
        :type sclass: :obj:`str`
        :returns: True if property name was changed
        :rtype: :obj:`bool`

        """
        ssvr = server.split("/")
        sclass = sclass or ssvr[0]
        if len(ssvr) == 1:
            server = "%s/*" % server
        res = False
        mss = self.db.get_server_list(server).value_string
        for ms in mss:
            devserv = self.db.get_device_class_list(ms).value_string
            dev = devserv[0::2]
            serv = devserv[1::2]
            for idx, ser in enumerate(serv):
                if ser == sclass:
                    if dev[idx]:
                        if not self.db.get_device_property(
                                dev[idx], newname)[newname]:
                            oldprop = self.db.get_device_property(
                                dev[idx], oldname)[oldname]
                            if oldprop:
                                oldprop = [p for p in oldprop if p]
                                self.db.put_device_property(
                                    dev[idx],
                                    {newname: oldprop})
                                self.db.delete_device_property(
                                    dev[idx], oldname)
                                res = True
        time.sleep(0.2)
        return res

    def changePropertyValue(self, server, newname, newvalue, sclass=None):
        """ changes/sets property value

        :param server: server name
        :type server: :obj:`str`
        :param newvalue: new property value
        :type newvalue: :obj:`str`
        :param newname: new property name
        :type newname: :obj:`str`
        :param sclass: server class name
        :type sclass: :obj:`str`
        :returns: True if property name was changed
        :rtype: :obj:`bool`

        """
        if "/" in server:
            sclass = sclass or server.split("/")[0]
            fserver = server
        else:
            sclass = sclass or server
            fserver = "%s/*" % server
        res = False
        value = json.loads(newvalue)
        mss = self.db.get_server_list(fserver).value_string
        for ms in mss:
            devserv = self.db.get_device_class_list(ms).value_string
            dev = devserv[0::2]
            serv = devserv[1::2]
            for idx, ser in enumerate(serv):
                if ser == sclass:
                    if dev[idx]:
                        if value:
                            value = [p for p in value if p]
                            self.db.put_device_property(
                                dev[idx],
                                {newname: value})
                            res = True
        time.sleep(0.2)
        return res

    def getStarterName(self, host):
        """ restarts server

        :param host: server host name
        :type host: :obj:`str`
        :returns: starter device name
        :rtype: :obj:`str`
        """
        admin = None
        if not host:
            host = socket.gethostname()
        admins = self.db.get_device_exported_for_class(
            "Starter").value_string
        eadmins = [adm for adm in admins
                   if self.db.get_device_exported(adm).value_string]
        if len(eadmins) == 1:
            admin = eadmins[0]
        elif len(eadmins) > 1:
            if 'tango/admin/%s' % host in eadmins:
                admin = 'tango/admin/%s' % host
            else:
                admin = eadmins[0]
        return admin

    def waitServerNotRunning(self, server=None, device=None,
                             adminproxy=None,
                             maxcnt=10000, verbose=True,
                             waitforproc=True):
        """  wait until device is exported and server is running

        :param server: server name, check if running when not None
        :type server: :obj:`str`
        :param device: device name, check if exported when not None
        :type device: :obj:`str`
        :param adminproxy: Starter device proxy
        :type adminproxy: :obj:`tango.DeviceProxy`
        :param maxcnt: maximal waiting time in 10ms
        :type maxcnt: :obj:`int`
        :param verbose: verbose mode
        :type verbose: :obj:`bool`
        :param waitforproc: wait for process list update
        :type waitforporc: :obj:`bool`
        :returns: True if server is running
        :rtype: :obj:`bool`
        """
        found = True
        cnt = 0
        running = True
        while found and cnt < maxcnt:
            try:
                if server and adminproxy is not None and running:
                    if verbose:
                        sys.stdout.write(".")
                        sys.stdout.flush()
                    adminproxy.UpdateServersInfo()
                    # running = adminproxy.RunningServers
                    running = adminproxy.DevGetRunningServers(True)
                    if server in running:
                        time.sleep(0.02)
                        cnt += 1
                        continue
                    else:
                        running = False
                if device:
                    if verbose:
                        sys.stdout.write(".")
                        sys.stdout.flush()
                    exl = self.db.get_device_exported(device)
                    if device in exl.value_string:
                        time.sleep(0.02)
                        cnt += 1
                        continue
                found = False
                if verbose:
                    if device or server:
                        print(" %s is not working" % (device or server))
            except Exception as e:
                print(str(e))
                time.sleep(0.02)
                found = True
            cnt += 1
        if waitforproc:
            time.sleep(1.5)
        return found

    def waitServerRunning(self, server=None, device=None,
                          adminproxy=None,
                          maxcnt=10000, verbose=True,
                          waitforproc=True):
        """  wait until device is exported and server is running

        :param server: server name, check if running when not None
        :type server: :obj:`str`
        :param device: device name, check if exported when not None
        :type device: :obj:`str`
        :param adminproxy: Starter device proxy
        :type adminproxy: :obj:`tango.DeviceProxy`
        :param maxcnt: maximal waiting time in 10ms
        :type maxcnt: :obj:`int`
        :param verbose: verbose mode
        :type verbose: :obj:`bool`
        :param waitforproc: wait for process list update
        :type waitforporc: :obj:`bool`
        :returns: True if server is running
        :rtype: :obj:`bool`
        """
        found = False
        cnt = 0
        exported = False
        while not found and cnt < maxcnt:
            try:
                if device and not exported:
                    if verbose:
                        sys.stdout.write(".")
                        sys.stdout.flush()
                    exl = self.db.get_device_exported(device)
                    if device not in exl.value_string:
                        time.sleep(0.02)
                        cnt += 1
                        continue
                    else:
                        exported = True
                if server is not None and adminproxy is not None:
                    if verbose:
                        sys.stdout.write(".")
                        sys.stdout.flush()
                    adminproxy.UpdateServersInfo()
                    # running = adminproxy.RunningServers
                    running = adminproxy.DevGetRunningServers(True)
                    if server not in running:
                        time.sleep(0.02)
                        cnt += 1
                        continue
                found = True
                if verbose:
                    if device or server:
                        print(" %s is working" % (device or server))
            except Exception as e:
                print(str(e))
                time.sleep(0.02)
                found = False
            cnt += 1
        if waitforproc:
            time.sleep(1.5)
        return found

    def killServer(self, server):
        sserver = server.split("/")
        if len(sserver) > 1:
            sname = sserver[0]
            instance = sserver[1]
        if sys.version_info > (3,):
            with subprocess.Popen(
                    "ps -ef | grep '%s %s' | grep -v grep"
                    % (sname, instance),
                    stdout=subprocess.PIPE, shell=True) as proc:

                pipe = proc.stdout
                res = str(pipe.read(), "utf8").split("\n")
                for r in res:
                    sr = r.split()
                    if len(sr) > 2:
                        subprocess.call(
                            "kill -9 %s" % sr[1], stderr=subprocess.PIPE,
                            shell=True)
                pipe.close()
        else:
            pipe = subprocess.Popen(
                "ps -ef | grep '%s %s' | grep -v grep"
                % (sname, instance),
                stdout=subprocess.PIPE, shell=True).stdout

            res = str(pipe.read()).split("\n")
            for r in res:
                sr = r.split()
                if len(sr) > 2:
                    subprocess.call(
                        "kill -9 %s" % sr[1], stderr=subprocess.PIPE,
                        shell=True)
            pipe.close()

    def restartServer(self, name, host=None, level=None,
                      restart=True, stopstart=True, wait=True,
                      timeout=None):
        """ restarts server

        :param name: server name
        :type name: :obj:`str`
        :param host: server host name
        :type host: :obj:`str`
        :param level: start up level
        :type level: :obj:`int`
        :param restart:  if server should be restarted
        :type restart: :obj:`bool`
        :param stopstart:  if server should be stopped and started
        :type stopstart: :obj:`bool`
        :param wait:  script should wait for the server
        :type wait: :obj:`bool`
        :param timeout: timeout for  start
        :type timeout: :obj:`float`
        """
        if name:
            admin = self.getStarterName(host)
            if not admin:
                raise Exception("Starter tango server is not running")
            if admin:
                servers = None
                started = None
                try:
                    adminproxy = tango.DeviceProxy(admin)
                    adminproxy.UpdateServersInfo()
                    servers = self.__registered_servers()
                    started = adminproxy.DevGetRunningServers(True)
                except Exception:
                    pass
                if servers:
                    for vl in set(servers):
                        svl = vl.split('\t')[0]
                        if name.startswith("NXSRecSelector") \
                                and svl.startswith("NXSRecSelector"):
                            self._changeLevel(svl, 4)
                        if (started and svl in started) or not restart:
                            if '/' in name:
                                cname = svl
                            else:
                                cname = svl.split('/')[0]
                            if cname == name:
                                if level is not None:
                                    self._changeLevel(
                                        svl, level, tohigher=False)
                                problems = False
                                if stopstart:
                                    if started and svl in started:
                                        try:
                                            # adminproxy.HardKillServer(svl)
                                            adminproxy.DevStop(svl)
                                            # self.killServer(svl)
                                        except Exception:
                                            adminproxy.HardKillServer(svl)
                                        problems = self.waitServerNotRunning(
                                            svl, None, adminproxy,
                                            verbose=None)
                                        if problems:
                                            self.killServer(svl)
                                            print("Server Running")
                                        sys.stdout.write(
                                            "Restarting: %s" % svl)
                                    else:
                                        sys.stdout.write("Starting: %s" % svl)
                                    sys.stdout.flush()
                                    problems = True
                                    counter = 0
                                    while problems and counter < 1000:
                                        try:
                                            sys.stdout.write('.')
                                            sys.stdout.flush()
                                            adminproxy.DevStart(svl)
                                            problems = False
                                        except Exception:
                                            counter += 1
                                            time.sleep(0.2)
                                if wait:
                                    try:
                                        maxcnt = int(float(timeout)/0.02)
                                    except Exception:
                                        maxcnt = 10000
                                    problems = not self.waitServerRunning(
                                        svl, None, adminproxy, maxcnt) \
                                        or problems
                                    if problems:
                                        print("%s was not restarted" % svl)
                                        print("Warning: "
                                              "Process with the server"
                                              "instance could be suspended")
                                else:
                                    print("")

    def __exported_servers(self):
        """ returns Servers for exported devices

        :rtype: :obj:`list` <:obj:`str`>
        :returns: server list
        """
        dvexported = self.db.get_device_exported("*").value_string
        return [sv for sv in self.db.get_server_list()
                if (set(self.db.get_device_class_list(sv).value_string[:2])
                    & set(dvexported))]

    def __registered_servers(self):
        """ returns registered Servers

        :rtype: :obj:`list` <:obj:`str`>
        :returns: server list
        """
        hosts = self.db.get_host_list().value_string
        svlist = []
        for ht in hosts:
            svlist.extend(self.db.get_host_server_list(ht).value_string)
        return svlist

    def stopServer(self, name, host=None):
        """ restarts server

        :param name: server name
        :type name: :obj:`str`
        :param host: server host name
        :type host: :obj:`str`
        """
        if name:
            admin = self.getStarterName(host)
            if not admin:
                raise Exception("Starter tango server is not running")
            if admin:
                servers = None
                started = None
                try:
                    adminproxy = tango.DeviceProxy(admin)
                    adminproxy.UpdateServersInfo()
                    started = adminproxy.DevGetRunningServers(True)
                    servers = self.__registered_servers()
                except Exception:
                    pass
                if servers:
                    for vl in set(servers):
                        svl = vl.split('\t')[0]
                        if started and svl in started:
                            if '/' in name:
                                cname = svl
                            else:
                                cname = svl.split('/')[0]
                            if cname == name:
                                if started and svl in started:
                                    try:
                                        adminproxy.DevStop(svl)
                                    except Exception:
                                        adminproxy.HardKillServer(svl)

                                    sys.stdout.write("Stopping: %s" % svl)
                                    sys.stdout.flush()
                                problems = self.waitServerNotRunning(
                                    svl, None, adminproxy)
                                print(" ")
                                if problems:
                                    print("%s was not stopped" % svl)
                                    print("Warning: Process with the server"
                                          "instance could be suspended")

    def _changeLevel(self, name, level, tohigher=True):
        """ change startup level

        :param name: server name
        :type name: :obj:`str`
        :param level: new startup level
        :type level: :obj:`int`
        :returns: True if level was changed
        :rtype: :obj:`bool`
        """
        sinfo = self.db.get_server_info(name)
        if not tohigher or level > sinfo.level:
            sinfo.level = level
        self.db.put_server_info(sinfo)
        return True

    def _startupServer(self, new, level, host, ctrl, device, postpone=False):
        """ starts the server up

        :param new: new server name
        :type new: :obj:`str`
        :param level: startup level
        :type level: :obj:`int`
        :param host: tango host name
        :type host: :obj:`str`
        :param ctrl: control mode
        :type ctrl: :obj:`str`
        :param device: device name
        :type device: :obj:`str`
        :param postpone: postpone start server
        :type postpone: :obj:`bool`
        :returns: True if server was started up
        :rtype: :obj:`bool`
        """
        server = self.db.get_server_class_list(new)
        if len(server) == 0:
            sys.stderr.write('Server ' + new.split('/')[0]
                             + ' not defined in database\n')
            sys.stderr.flush()
            return False
        admin = self.getStarterName(host)
        if not admin:
            raise Exception("Starter tango server is not running")
        adminproxy = tango.DeviceProxy(admin)
        startdspaths = self.db.get_device_property(
            admin,
            "StartDsPath")["StartDsPath"]
        if '/usr/bin' not in startdspaths:
            if startdspaths:
                startdspaths = [p for p in startdspaths if p]
            else:
                startdspaths = []
            startdspaths.append('/usr/bin')
            self.db.put_device_property(
                admin, {"StartDsPath": startdspaths})
            adminproxy.Init()

        sinfo = self.db.get_server_info(new)
        sinfo.name = new
        sinfo.host = host
        sinfo.mode = ctrl
        sinfo.level = level
        self.db.put_server_info(sinfo)
        adminproxy.UpdateServersInfo()
        running = adminproxy.DevGetRunningServers(True)
        # running = adminproxy.RunningServers
        if new not in running and not postpone:
            adminproxy.DevStart(new)
        else:
            print("%s NOT STARTED" % new)

        adminproxy.UpdateServersInfo()

        if not postpone:
            sys.stdout.write("waiting for server ")
            sys.stdout.flush()
            return self.waitServerRunning(new, device, adminproxy)

    def createDataWriter(self, beamline, masterhost, postpone=False):
        """ creates data writer

        :param beamline: beamline name
        :type beamline: :obj:`str`
        :param masterhost: master host of data writer
        :type masterhost: :obj:`str`
        :param postpone: postpone starting flag
        :type postpone: :obj:`bool`
        :returns: True if server was created
        :rtype: :obj:`bool`
        """
        if not beamline:
            print("createDataWriter: no beamline given ")
            return False
        if not masterhost:
            print("createDataWriter: no masterhost given ")
            return False

        class_name = 'NXSDataWriter'
        server = class_name
        server_name = server + '/' + masterhost
        full_class_name = 'NXSDataWriter/' + masterhost
        self.writer_name = "%s/nxsdatawriter/%s" % (beamline, masterhost)
        if server_name not in self.db.get_server_list(server_name):
            print("creating: %s" % server_name)

            if server_name in self.db.get_server_list(server_name):
                print("createDataWriter: DB contains already %s" % server_name)
                return False

            di = tango.DbDevInfo()
            di.name = self.writer_name
            di._class = class_name
            di.server = server_name

            self.db.add_device(di)
            self.db.put_device_property(self.writer_name,
                                        {'NumberOfThreads': 100})

        elif (self.writer_name not in
              self.db.get_device_class_list(server_name).value_string):
            print("\ncreateDataWriter: %s already exists. "
                  "To change its device name please remove it." % server_name)
            return False

        hostname = socket.gethostname()

        self._startupServer(full_class_name, 1, hostname, 1, self.writer_name,
                            postpone)

        return True

    def createServer(self, server_name, beamline, masterhost, hostname=None,
                     class_name=None, jsondeviceproperties=None,
                     postpone=False):
        """ creates tango server in DB

        :param server_name: server name
        :type server_name: :obj:`str`
        :param beamline: beamline name
        :type beamline: :obj:`str`
        :param masterhost: master host of data writer
        :type masterhost: :obj:`str`
        :param hostname: host to run the config server
        :type hostname: :obj:`str`
        :param class_name: class name
        :type clas_name: :obj:`str`
        :param jsondeviceproperties: json device properties
        :type jsondeviceproperties: :obj:`str`
        :param postpone: postpone starting flag
        :type postpone: :obj:`bool`
        :returns: True if server was created
        :rtype: :obj:`bool`
        """
        if not beamline:
            print("createServer: no beamline given ")
            return False
        if not masterhost:
            print("createServer: no masterhost given ")
            return False

        if "/" not in server_name:
            server_name = server_name + '/' + masterhost
        server = server_name.split("/")[0]
        if not class_name:
            class_name = server
        self.device_name = "%s/%s/%s" % (beamline, server.lower(), masterhost)
        if server_name not in self.db.get_server_list(server_name):
            print("creating: %s" % server_name)

            if server_name in self.db.get_server_list(server_name):
                print("createDataWriter: DB contains already %s" % server_name)
                return False

            try:
                properties = dict(json.loads(jsondeviceproperties))
            except Exception:
                print("\ncreateServer: properties cannot be load")
                properties = {}
            if not properties:
                properties = {self.device_name: {}}

            print("PROP", properties)
            for dv, props in properties.items():
                di = tango.DbDevInfo()
                di.name = dv
                di._class = class_name
                di.server = server_name
                self.db.add_device(di)
                try:
                    self.db.put_device_property(dv, props)
                except Exception:
                    print("\ncreateServer: properties cannot be added")

        elif (self.device_name not in
              self.db.get_device_class_list(server_name).value_string):
            print("\ncreateServer: %s already exists. "
                  "To change its device name please remove it." % server_name)
            return False

        if not hostname:
            hostname = socket.gethostname()

        self._startupServer(server_name, 1, hostname, 1, self.device_name,
                            postpone)

        return True

    def createConfigServer(self, beamline, masterhost, jsonsettings=None,
                           hostname=None, postpone=False):
        """ creates configuration server

        :param beamline: beamline name
        :type beamline: :obj:`str`
        :param masterhost: master host of data writer
        :type masterhost: :obj:`str`
        :param jsonsettings: connection settings to DB in json
        :type jsonsettings: :obj:`str`
        :param hostname: host to run the config server
        :type hostname: :obj:`str`
        :param postpone: postpone starting flag
        :type postpone: :obj:`bool`
        :returns: True if server was created
        :rtype: :obj:`bool`
        """
        if not beamline:
            print("createConfigServer: no beamline given ")
            return False
        if not masterhost:
            print("createConfigServer: no masterhost given ")
            return False

        class_name = 'NXSConfigServer'
        server = class_name
        server_name = server + '/' + masterhost
        self.cserver_name = "%s/nxsconfigserver/%s" % (beamline, masterhost)
        if server_name not in self.db.get_server_list(server_name):
            print("creating: %s" % server_name)

            if server_name in self.db.get_server_list(server_name):
                print("createConfigServer: DB contains already %s"
                      % server_name)
                return False

            di = tango.DbDevInfo()
            di.name = self.cserver_name
            di._class = class_name
            di.server = server_name

            self.db.add_device(di)
            self.db.put_device_property(
                self.cserver_name, {'VersionLabel': '%s@%s' % (
                    beamline.upper(), masterhost.upper())})
        elif (self.cserver_name not in
              self.db.get_device_class_list(server_name).value_string):
            print("\ncreateConfigServer: %s already exists. "
                  "To change its device name please remove it." % server_name)
            return False

        if not hostname:
            hostname = self.db.get_db_host().split(".")[0]

        self._startupServer(server_name, 1, hostname, 1, self.cserver_name,
                            postpone)

        if not postpone:
            dp = tango.DeviceProxy(self.cserver_name)
            if dp.state() != tango.DevState.ON:
                dp.Close()
            if jsonsettings:
                dp = tango.DeviceProxy(self.cserver_name)
                dp.JSONSettings = jsonsettings
            try:
                dp.Open()
            except Exception:
                try:
                    jsettings = json.loads(jsonsettings)
                    jsettings['read_default_file'] = \
                        '/var/lib/nxsconfigserver/.my.cnf'
                    dp.JSONSettings = str(json.dumps(jsettings))
                    dp.Open()
                except Exception:
                    try:
                        jsettings['read_default_file'] = \
                            '/var/lib/nxsconfigserver/.my.cnf'
                        dp.JSONSettings = str(json.dumps(jsettings))
                        dp.Open()
                    except Exception:
                        print("createConfigServer: "
                              "%s cannot connect the"
                              " database with JSONSettings: \n%s " % (
                                  self.cserver_name, jsonsettings))
                        print("try to change the settings")
                        return False

        return True

    def createSelector(self, beamline, masterhost, writer=None, cserver=None,
                       postpone=False):
        """ creates selector server

        :param beamline: beamline name
        :type beamline: :obj:`str`
        :param masterhost: master host of data writer
        :type masterhost: :obj:`str`
        :param writer: writer device name
        :type writer: :obj:`str`
        :param cserver: configuration server device name
        :type cserver: :obj:`str`
        :param postpone: postpone starting flag
        :type postpone: :obj:`bool`
        :returns: True if server was created
        :rtype: :obj:`bool`
        """
        if not beamline:
            print("createSelector: no beamline given ")
            return False
        if not masterhost:
            print("createSelector: no masterhost given ")
            return False
        if writer:
            self.writer_name = writer
        if cserver:
            self.cserver_name = cserver

        class_name = 'NXSRecSelector'
        server = class_name
        server_name = server + '/' + masterhost
        full_class_name = 'NXSRecSelector/' + masterhost
        device_name = "%s/nxsrecselector/%s" % (beamline, masterhost)
        if server_name not in self.db.get_server_list(server_name):
            print("creating: %s" % server_name)

            if server_name in self.db.get_server_list(server_name):
                print("createSelector: DB contains already %s" % server_name)
                return False

            di = tango.DbDevInfo()
            di.name = device_name
            di._class = class_name
            di.server = server_name
            self.db.add_device(di)

        elif (device_name not in
              self.db.get_device_class_list(server_name).value_string):
            print("\ncreateSelector: %s already exists. "
                  "To change its device name please remove it." % server_name)
            return False

        hostname = socket.gethostname()

        self._startupServer(full_class_name, 4, hostname, 1,
                            device_name, postpone)

        if not postpone:
            if self.writer_name or self.cserver_name:
                dp = tango.DeviceProxy(device_name)
                dp.ping()
                self.waitServerRunning(None, device_name)
                if self.cserver_name:
                    dp.configDevice = self.cserver_name
                if self.writer_name:
                    dp.writerDevice = self.writer_name
        return True


class Set(Runner):

    """ set runner"""

    #: (:obj:`str`) command description
    description = "set up NXSConfigServer NXSDataWriter " \
                  + "and NXSRecSelector servers"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxsetup set\n" \
        + "       nxsetup set -b p09 -m haso228 -u p09user " \
        + "-d nxsconfig NXSConfigServer\n" \
        + "       nxsetup set NexusWriter/haso228  -k NexusWriter  " \
        + "-y '{\"p00/bliss_nexuswriter/test_session\":" \
        + "{\"session\":\"test_session\"," \
        + "\"beacon_host\":\"haso228:25000\"}}'  -t\n" \
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            "-b", "--beamline", action="store",
            dest="beamline", help="name of the beamline"
            " ( default: 'nxs' )")
        parser.add_argument(
            "-m", "--masterhost", action="store",
            dest="masterhost", help="the host that stores the Mg"
            " ( default: <localhost> )")
        parser.add_argument(
            "-c", "--confighost", action="store",
            dest="confighost", help="the host run the config server"
            " ( default: <mysqlhost> )")
        parser.add_argument(
            "-r", "--runhost", action="store",
            dest="runhost", help="the host to run the server"
            " ( default: localhost )")
        parser.add_argument(
            "-u", "--user", action="store",
            dest="user", help="the local user"
            " ( default: 'tango' )")
        parser.add_argument(
            "-d", "--database", action="store",
            dest="dbname", help="the database name"
            "  ( default: 'nxsconfig')")
        parser.add_argument(
            "-j", "--csjson", action="store",
            dest="csjson",
            help="JSONSettings for the configuration server. "
            "( default: '{\"host\": \"localhost\",\"db\": <DBNAME>,"
            " \"use_unicode\": true,"
            " \"read_default_file\": <MY_CNF_FILE>}'"
            "  where <MY_CNF_FILE> stays for"
            " \"/home/<USER>/.my.cnf\""
            " or \"/var/lib/nxsconfigserver/.my.cnf\" )")
        parser.add_argument(
            "-k", "--class-name", action="store",
            dest="classname", help="tango server class name")
        parser.add_argument(
            "-y", "--json-device-properties", action="store",
            dest="propjson",
            help="JSON tango device properties "
            "( default: '{}' )")
        parser.add_argument(
            "-t", "--postpone", action="store_true",
            default=False, dest="postpone",
            help="do not start the server")
        parser.add_argument(
            'args', metavar='server_name',
            type=str, nargs='*',
            help='server names, e.g.: NXSRecSelector NXSConfigServer')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        local_user = None
        args = options.args or []
        if os.path.isfile('/home/etc/local_user'):
            with open('/home/etc/local_user') as fl:
                local_user = fl.readline()
        elif _hostname in knownHosts.keys():
            local_user = knownHosts["user"]

        if options.beamline is None:
            if _hostname in knownHosts.keys():
                options.beamline = knownHosts[_hostname]['beamline']
            else:
                options.beamline = 'nxs'

        if options.masterhost is None:
            if _hostname in knownHosts.keys():
                options.masterhost = knownHosts[_hostname]['masterhost']
            else:
                options.masterhost = _hostname

        if options.user is None:
            if local_user:
                options.user = local_user
            else:
                options.user = 'tango'

        if options.dbname is None:
            if _hostname in knownHosts.keys():
                options.dbname = knownHosts[_hostname]['dbname']
            else:
                options.dbname = 'nxsconfig'

        print("\noptions are set to:  -b %s -m %s -u %s -d %s \n" % (
            options.beamline,
            options.masterhost,
            options.user,
            options.dbname,
        ))

        setUp = SetUp()

        if not args or "NXSDataWriter" in args:
            if not setUp.createDataWriter(
                    beamline=options.beamline,
                    masterhost=options.masterhost):
                print("startup failed to create the nexus data writer")
                sys.exit(255)

        if options.csjson:
            jsonsettings = options.csjson
        else:
            jsonsettings = '{"db":"%s",' % options.dbname \
                + ' "read_default_file":"/home/%s/.my.cnf",' % options.user \
                + ' "use_unicode":true}'

        if options.propjson:
            jsondeviceproperties = options.propjson
        else:
            jsondeviceproperties = '{}'

        if not args or "NXSConfigServer" in args:
            if not setUp.createConfigServer(
                    beamline=options.beamline,
                    masterhost=options.masterhost,
                    jsonsettings=jsonsettings,
                    hostname=options.confighost,
                    postpone=options.postpone):
                print("startup failed to create the nexus config server")
                sys.exit(255)

        if not args or "NXSRecSelector" in args:
            if not setUp.createSelector(
                    beamline=options.beamline,
                    masterhost=options.masterhost,
                    postpone=options.postpone):
                print("startup failed to create the nexus selector server")
                sys.exit(255)
        for arg in args:
            if arg not in ["NXSDataWriter", "NXSRecSelector",
                           "NXSConfigServer"]:
                if not setUp.createServer(
                        arg,
                        beamline=options.beamline,
                        masterhost=options.masterhost,
                        hostname=options.runhost,
                        class_name=options.classname,
                        jsondeviceproperties=jsondeviceproperties,
                        postpone=options.postpone):
                    print("startup failed to create the %s server" % arg)
                    sys.exit(255)


class Start(Runner):

    """ start runner"""

    #: (:obj:`str`) command description
    description = "start tango server"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxsetup start Pool/haso228 -l 2\n" \
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            "-l", "--level", action="store", type=int, default=-1,
            dest="level", help="startup level")
        parser.add_argument(
            "-z", "--timeout", action="store", type=float, default=None,
            dest="timeout", help="timeout in seconds")
        parser.add_argument(
            "-e", "--no-wait", action="store_true",
            default=False, dest="nowait",
            help="do not wait")
        parser.add_argument(
            'args', metavar='server_name',
            type=str, nargs='*',
            help='server names, e.g.: NXSRecSelector NXSDataWriter/TDW1')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        args = options.args or []
        setUp = SetUp()
        servers = args if args else [
            "NXSConfigServer", "NXSRecSelector", "NXSDataWriter"]
        for server in servers:
            setUp.restartServer(
                server,
                level=(options.level if options.level > -1 else None),
                restart=False, timeout=options.timeout,
                wait=(not options.nowait))


class Wait(Runner):

    """ start runner"""

    #: (:obj:`str`) command description
    description = "wait for tango server"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxsetup wait Pool/haso228 -z 30\n" \
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            "-z", "--timeout", action="store", type=float, default=None,
            dest="timeout", help="timeout in seconds")
        parser.add_argument(
            'args', metavar='server_name',
            type=str, nargs='*',
            help='server names, e.g.: NXSRecSelector NXSDataWriter/TDW1')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        args = options.args or []
        setUp = SetUp()
        servers = args if args else [
            "NXSConfigServer", "NXSRecSelector", "NXSDataWriter"]
        for server in servers:
            setUp.restartServer(
                server,
                restart=False, stopstart=False, timeout=options.timeout)


class MoveProp(Runner):

    """ move-prop runner"""

    #: (:obj:`str`) command description
    description = "change property name"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxsetup move-prop -n DefaultPreselectedComponents" \
        + " -o DefaultAutomaticComponents NXSRecSelector\n" \
        + "       nxsetup move-prop -t -n DefaultPreselectedComponents " \
        + " -o DefaultAutomaticComponents NXSRecSelector\n" \
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            "-n", "--newname", action="store",
            dest="newname", help="(new) property name")
        parser.add_argument(
            "-o", "--oldname", action="store",
            dest="oldname", help="old property name")
        parser.add_argument(
            "-t", "--postpone", action="store_true",
            default=False, dest="postpone",
            help="do not restart the server")
        parser.add_argument(
            "-z", "--timeout", action="store", type=float, default=None,
            dest="timeout", help="timeout in seconds")
        parser.add_argument(
            "-e", "--no-wait", action="store_true",
            default=False, dest="nowait",
            help="do not wait")
        parser.add_argument(
            'args', metavar='server_name',
            type=str, nargs='*',
            help='server names, e.g.: NXSRecSelector NXSDataWriter/TDW1')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        parser = self._parser
        args = options.args or []
        if not options.newname or not options.oldname or not args:
            parser.print_help()
            sys.exit(255)
        servers = args or []
        setUp = SetUp()
        for server in servers:
            if setUp.changePropertyName(
                    server, options.oldname, options.newname):
                if not options.postpone:
                    setUp.restartServer(
                        server, timeout=options.timeout,
                        wait=(not options.nowait))


class ChangeProp(Runner):

    """ change-prop runner"""

    #: (:obj:`str`) command description
    description = "change property value"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxsetup change-prop -n ClientRecordKeys -t -w " \
        + "\"[\\\"phoibos_scan_command\\\",\\\"phoibos_scan_comment\\\"]\" " \
        + "NXSRecSelector/r228\n" \
        + "       nxsetup change-prop -n DefaultPreselectedComponents -w " \
        + "\"[\\\"pinhole1\\\",\\\"slit2\\\"]\" NXSRecSelector/r228\n" \
        + "       nxsetup change-prop -n StartDsPath -w " \
        + "\"[\\\"/usr/bin\\\",\\\"/usr/lib/tango\\\"]\" Starter\n" \
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            "-n", "--newname", action="store",
            dest="newname", help="(new) property name")
        parser.add_argument(
            "-w", "--propvalue", action="store",
            dest="propvalue", help="new property value")
        parser.add_argument(
            "-t", "--postpone", action="store_true",
            default=False, dest="postpone",
            help="do not restart the server")
        parser.add_argument(
            "-z", "--timeout", action="store", type=float, default=None,
            dest="timeout", help="timeout in seconds")
        parser.add_argument(
            "-e", "--no-wait", action="store_true",
            default=False, dest="nowait",
            help="do not wait")
        parser.add_argument(
            'args', metavar='server_name',
            type=str, nargs='*',
            help='server names, e.g.: NXSRecSelector NXSDataWriter/TDW1')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        parser = self._parser
        args = options.args or []
        if not options.newname or not options.propvalue or not args:
            parser.print_help()
            sys.exit(255)
        servers = args or []
        setUp = SetUp()
        for server in servers:
            if setUp.changePropertyValue(
                    server, options.newname, options.propvalue):
                if not options.postpone:
                    setUp.restartServer(
                        server,
                        timeout=options.timeout, wait=(not options.nowait))


class AddRecorderPath(Runner):

    """ add-recorder-path runner"""

    #: (:obj:`str`) command description
    description = "add-recorder-path into MacroServer(s) property"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxsetup add-recorder-path "\
        + "/usr/share/pyshared/sardananxsrecorder\n" \
        + "       nxsetup add-recorder-path -t "\
        + "/usr/share/pyshared/sardananxsrecorder\n" \
        + "       nxsetup add-recorder-path "\
        + "/usr/share/pyshared/sardananxsrecorder -i haso\n" \
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            "-t", "--postpone", action="store_true",
            default=False, dest="postpone",
            help="do not restart the server")
        parser.add_argument(
            "-z", "--timeout", action="store", type=float, default=None,
            dest="timeout", help="timeout in seconds")
        parser.add_argument(
            "-e", "--no-wait", action="store_true",
            default=False, dest="nowait",
            help="do not wait")
        parser.add_argument(
            "-i", "--instance", action="store",
            dest="instance",
            help="macroserver instance name, i.e. haso"
            " ( default: '*')")

    def postauto(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            'recpath', metavar='recorder_path',
            type=str, nargs=1,
            help='sardana recorder path')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        if options.instance is None:
            options.instance = "*"

        setUp = SetUp()
        if setUp.changeRecorderPath(
                options.recpath[0], options.instance):
            if not options.postpone:
                if options.instance != "*":
                    ms = "MacroServer/%s" % options.instance
                else:
                    ms = "MacroServer"
                setUp.restartServer(ms, timeout=options.timeout,
                                    wait=(not options.nowait))


class Restart(Runner):

    """ restart runner"""

    #: (:obj:`str`) command description
    description = "restart tango server"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxsetup restart Pool/haso228 -l 2\n" \
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            "-l", "--level", action="store", type=int, default=-1,
            dest="level", help="startup level")
        parser.add_argument(
            "-z", "--timeout", action="store", type=float, default=None,
            dest="timeout", help="timeout in seconds")
        parser.add_argument(
            "-e", "--no-wait", action="store_true",
            default=False, dest="nowait",
            help="do not wait")
        parser.add_argument(
            'args', metavar='server_name',
            type=str, nargs='*',
            help='server names, e.g.: NXSRecSelector NXSDataWriter/TDW1')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        args = options.args or []
        setUp = SetUp()
        servers = args if args else [
            "NXSConfigServer", "NXSRecSelector", "NXSDataWriter"]
        for server in servers:
            setUp.restartServer(
                server, level=(options.level if options.level > -1 else None),
                timeout=options.timeout, wait=(not options.nowait))


class Stop(Runner):

    """ Stop runner"""

    #: (:obj:`str`) command description
    description = "stop tango server"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxsetup stop Pool/haso228 \n" \
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            'args', metavar='server_name',
            type=str, nargs='*',
            help='server names, e.g.: NXSRecSelector NXSDataWriter/TDW1')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        args = options.args or []
        setUp = SetUp()
        servers = args if args else [
            "NXSConfigServer", "NXSRecSelector", "NXSDataWriter"]
        for server in servers:
            setUp.stopServer(server)


def _supportoldcommands():
    """ replace the old command names to the new ones
    """

    oldnew = {
        '-x': 'set',
        '--execute': 'set',
        '-r': 'restart',
        '--restart': 'restart',
        '-s': 'start',
        '--start': 'start',
        '-p': 'move-prop',
        '--move-prop': 'move-prop',
        '-c': 'change-prop',
        '--change-prop': 'change-prop',
        '-a': 'add-recorder-path',
        '--add-recorder-path': 'add-recorder-path',
    }

    if sys.argv and len(sys.argv) > 1:
        if sys.argv[1] in oldnew.keys():
            sys.argv[1] = oldnew[sys.argv[1]]


def main():
    """ the main program function
    """
    description = "Command-line tool for setting up  NeXus Tango Servers"

    _supportoldcommands()

    epilog = 'For more help:\n  nxsetup <sub-command> -h'
    if _hostname in knownHosts.keys():
        local_user = None
        if os.path.isfile('/home/etc/local_user'):
            with open('/home/etc/local_user') as fl:
                local_user = fl.readline()
        elif _hostname in knownHosts.keys():
            local_user = knownHosts["user"]
        epilog += "\n\n  (%s is known: -b %s -m %s -u %s -d %s )" % (
            _hostname,
            knownHosts[_hostname]['beamline'],
            knownHosts[_hostname]['masterhost'],
            local_user,
            knownHosts[_hostname]['dbname']
        )

    parser = NXSArgParser(
        description=description, epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.cmdrunners = [
        ('set', Set),
        ('restart', Restart),
        ('start', Start),
        ('stop', Stop),
        ('wait', Wait),
        ('move-prop', MoveProp),
        ('change-prop', ChangeProp),
        ('add-recorder-path', AddRecorderPath),
    ]
    runners = parser.createSubParsers()

    try:
        options = parser.parse_args()
    except ErrorException as e:
        sys.stderr.write("Error: %s\n" % str(e))
        sys.stderr.flush()
        parser.print_help()
        print("")
        sys.exit(255)

    if options.subparser is None:
        sys.stderr.write(
            "Error: %s\n" % str("too few arguments"))
        sys.stderr.flush()
        parser.print_help()
        print("")
        sys.exit(255)

    runners[options.subparser].run(options)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
# \package test nexdatas
# \file XMLConfiguratorTest.py
# unittests for field Tags running Tango Server
#
import unittest
import os
import sys
import random
import struct
import binascii
from nxstools import nxscreate


try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO


if sys.version_info > (3,):
    unicode = str
    long = int


# if 64-bit machione
IS64BIT = (struct.calcsize("P") == 8)

# from nxsconfigserver.XMLConfigurator  import XMLConfigurator
# from nxsconfigserver.Merger import Merger
# from nxsconfigserver.Errors import (
# NonregisteredDBRecordError, UndefinedTagError,
#                                    IncompatibleNodeError)
# import nxsconfigserver


def myinput(w, text):
    myio = os.fdopen(w, 'w')
    myio.write(text)

    # myio.close()


# test fixture
class NXSCreateTest(unittest.TestCase):

    # constructor
    # \param methodName name of the test method
    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

        self.helperror = "Error: too few arguments\n"

        self.helpinfo = """usage: nxscreate [-h]
                 {clientds,tangods,deviceds,onlinecp,onlineds,poolds,stdcomp,comp,secopcp,compare}
                 ...

 Command-line tool for creating NXSConfigServer configuration of Nexus Files

positional arguments:
  {clientds,tangods,deviceds,onlinecp,onlineds,poolds,stdcomp,comp,secopcp,compare}
                        sub-command help
    clientds            create client datasources
    tangods             create tango datasources
    deviceds            create datasources for all device attributes
    onlinecp            create component from online.xml file
    onlineds            create datasources from online.xml file
    poolds              create datasources from sardana pool device
    stdcomp             create component from the standard templates
    comp                create simple components
    secopcp             create secop components
    compare             compare two online.xml files

optional arguments:
  -h, --help            show this help message and exit

For more help:
  nxscreate <sub-command> -h

"""

        try:
            # random seed
            self.seed = long(binascii.hexlify(os.urandom(16)), 16)
        except NotImplementedError:
            import time
            # random seed
            self.seed = long(time.time() * 256)  # use fractional seconds

        self.__rnd = random.Random(self.seed)

        self._bint = "int64" if IS64BIT else "int32"
        self._buint = "uint64" if IS64BIT else "uint32"
        self._bfloat = "float64" if IS64BIT else "float32"

        self.__args = '{"db":"nxsconfig", ' \
                      '"read_default_file":"/etc/my.cnf", "use_unicode":true}'
        self.__cmps = []
        self.__ds = []
        self.__man = []
        self.maxDiff = None
        self.children = ("record", "doc", "device", "database", "query",
                         "datasource", "result")

        from os.path import expanduser
        home = expanduser("~")
        self.__args2 = '{"db":"nxsconfig", ' \
                       '"read_default_file":"%s/.my.cnf", ' \
                       '"use_unicode":true}' % home

    # test starter
    # \brief Common set up
    def setUp(self):
        print("\nsetting up...")
        print("SEED = %s" % self.seed)

    # test closer
    # \brief Common tear down
    def tearDown(self):
        print("tearing down ...")

    # Exception tester
    # \param exception expected exception
    # \param method called method
    # \param args list with method arguments
    # \param kwargs dictionary with method arguments
    def myAssertRaise(self, exception, method, *args, **kwargs):
        try:
            error = False
            method(*args, **kwargs)
        except Exception:
            error = True
        self.assertEqual(error, True)

    # sets selection configuration
    # \param selectionc configuration instance
    # \param selection selection configuration string
    def setSelection(self, selectionc, selection):
        selectionc.selection = selection

    # gets selectionconfiguration
    # \param selectionc configuration instance
    # \returns selection configuration string
    def getSelection(self, selectionc):
        return selectionc.selection

    # comp_available test
    # \brief It tests XMLConfigurator
    def test_default(self):
        fun = sys._getframe().f_code.co_name
        print("Run: %s.%s() " % (self.__class__.__name__, fun))

        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = mystdout = StringIO()
        sys.stderr = mystderr = StringIO()
        old_argv = sys.argv
        sys.argv = ['nxscreate']
        with self.assertRaises(SystemExit):
            nxscreate.main()

        sys.argv = old_argv
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        vl = mystdout.getvalue()
        er = mystderr.getvalue()
        self.assertEqual(
            "".join(self.helpinfo.split()).replace(
                "optionalarguments:", "options:"),
            "".join(vl.split()).replace("optionalarguments:", "options:"))
        self.assertEqual(self.helperror, er)

    # comp_available tesQt
    # \brief It tests XMLConfigurator
    def test_help(self):
        fun = sys._getframe().f_code.co_name
        print("Run: %s.%s() " % (self.__class__.__name__, fun))

        helps = ['-h', '--help']
        for hl in helps:
            old_stdout = sys.stdout
            old_stderr = sys.stderr
            sys.stdout = mystdout = StringIO()
            sys.stderr = mystderr = StringIO()
            old_argv = sys.argv
            sys.argv = ['nxscreate', hl]
            with self.assertRaises(SystemExit):
                nxscreate.main()

            sys.argv = old_argv
            sys.stdout = old_stdout
            sys.stderr = old_stderr
            vl = mystdout.getvalue()
            er = mystderr.getvalue()
            self.assertEqual(
                "".join(self.helpinfo.split()).replace(
                    "optionalarguments:", "options:"),
                "".join(vl.split()).replace("optionalarguments:", "options:"))
            self.assertEqual('', er)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

"""  pyeval helper functions for pco """


def postrun(commonblock,
            filestartnum,
            filedir,
            nbframes,
            filepostfix,
            fileprefix,
            filestartnum_str):
    """ code for postrun  datasource

    :param commonblock: commonblock of nxswriter
    :type commonblock: :obj:`dict`<:obj:`str`, `any`>
    :param filestartnum:  file start number
    :type filestartnum: :obj:`int`
    :param filedir: file directorentry name
    :type filedir: :obj:`str`
    :param nbframes: number of frams
    :type nbframes: :obj:`int`
    :param filepostfix: file postfix
    :type filepostfix: :obj:`str`
    :param fileprefix: file prefix
    :type fileprefix: :obj:`str`
    :param filestartnum_str: filestartnum string
    :type filestartnum_str: :obj:`str`
    :returns: postrun string
    :rtype: :obj:`str`
    """
    unixdir = filedir.replace("\\", "/")
    if len(unixdir) > 1 and unixdir[1] == ":":
        unixdir = "/data" + unixdir[2:]
    if unixdir and unixdir[-1] == "/":
        unixdir = unixdir[:-1]
    filestartnumer = commonblock[filestartnum_str] - 1
    result = "" + unixdir + "/" + fileprefix + "%05d"
    result += filepostfix + ":"
    filelastnumber = filestartnum - 1
    if "__root__" in commonblock.keys():
        root = commonblock["__root__"]
        if hasattr(root, "currentfileid") and hasattr(root, "stepsperfile"):
            spf = root.stepsperfile
            cfid = root.currentfileid
            if spf > 0 and cfid > 0:
                filelastnumber = min(
                    filestartnumer + cfid * nbframes * spf - 1, filelastnumber)
                filestartnum = filestartnumer + \
                    (cfid - 1) * nbframes * spf
    result += str(filestartnumer) + ":" + str(filelastnumber)
    return result



#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

""" NeXus main metadata viewer """

from . import filewriter
import fnmatch
import json
import sys
import xml.etree.ElementTree as et
import numpy as np
import math
import time
import dateutil.parser
import re
from lxml.etree import XMLParser

from nxstools.nxsparser import ParserTools


class numpyEncoder(json.JSONEncoder):
    """ numpy json encoder with list
    """

    def default(self, obj):
        """ default encoder

        :param obj: numpy array object
        :type obj: :obj:`object` or `any`
        """
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, bytes):
            try:
                return obj.decode("utf-8")
            except Exception:
                return obj.decode()
        return json.JSONEncoder.default(self, obj)


class numpyEncoderNull(numpyEncoder):
    """ numpy json encoder with list with nan/inf to null
    """
    def encode(self, obj, *args, **kwargs):
        return numpyEncoder.encode(self, infNaN2None(obj), *args, **kwargs)


def infNaN2None(obj):
    """ replace inf and NaN to None
    """
    if isinstance(obj, dict):
        return {ky: infNaN2None(vl) for ky, vl in obj.items()}
    elif isinstance(obj, list):
        return [infNaN2None(it) for it in obj]
    elif isinstance(obj, float) and math.isinf(obj):
        return None
    elif isinstance(obj, float) and math.isnan(obj):
        return None
    return obj


_regex = r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-' \
    r'(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])' \
    r'(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$'

_match_iso8601 = re.compile(_regex).match


def isoDate(text):
    """ convert date to iso format

    :param text: date text to convert
    :type text: :obj:`str`
    :returns: date in iso format
    :rtype: :obj:`str`
    """
    result = ""
    try:
        try:
            if _match_iso8601(text) is not None:
                result = text
        except Exception:
            pass
        if not result:
            date = dateutil.parser.parse(text)
            fmt = '%Y-%m-%dT%H:%M:%S.%f%z'
            if date.tzinfo is None:
                tzone = time.tzname[0]
                try:
                    if sys.version_info >= (3, 9):
                        import zoneinfo
                        tz = zoneinfo.ZoneInfo(tzone)
                        date = date.replace(tzinfo=tz)
                    else:
                        import pytz
                        tz = pytz.timezone(tzone)
                        date = tz.localize(date)
                except Exception:
                    import tzlocal
                    tz = tzlocal.get_localzone()
                    date = date.replace(tzinfo=tz)
            result = str(date.strftime(fmt))
    except Exception:
        result = text
    return result


def getdsname(xmlstring):
    """ provides datasource name from datasource xml string

    :param xmlstring: datasource xml string
    :type xmlstring: :obj:`str`
    """

    if sys.version_info > (3,):
        node = et.fromstring(
            bytes(xmlstring, "UTF-8"),
            parser=XMLParser(collect_ids=False))
    else:
        node = et.fromstring(
            xmlstring,
            parser=XMLParser(collect_ids=False))
    if node.tag == 'datasource':
        nodes = [node]
    else:
        nodes = node.findall(".//datasource")
    dsname = ""
    if nodes and "name" in nodes[0].attrib:
        dsname = nodes[0].attrib["name"]
    return dsname or ""


def getdstype(xmlstring):
    """ provides datasource type from datasource xml string

    :param xmlstring: datasource xml string
    :type xmlstring: :obj:`str`
    """

    if sys.version_info > (3,):
        node = et.fromstring(
            bytes(xmlstring, "UTF-8"),
            parser=XMLParser(collect_ids=False))
    else:
        node = et.fromstring(
            xmlstring,
            parser=XMLParser(collect_ids=False))
    if node.tag == 'datasource':
        nodes = [node]
    else:
        nodes = node.findall(".//datasource")
    dstype = ""
    if nodes:
        dstype = nodes[0].attrib["type"]
    return dstype


def getdssource(xmlstring):
    """ provides source from datasource xml string

    :param xmlstring: datasource xml string
    :type xmlstring: :obj:`str`
    """

    if sys.version_info > (3,):
        node = et.fromstring(
            bytes(xmlstring, "UTF-8"),
            parser=XMLParser(collect_ids=False))
    else:
        node = et.fromstring(
            xmlstring,
            parser=XMLParser(collect_ids=False))
    if node.tag == 'datasource':
        nodes = [node]
    else:
        nodes = node.findall(".//datasource")
    dssource = ""
    if nodes:
        ds = nodes[0]
        dssource = ParserTools.getRecord(ds)
    return dssource


class NXSFileParser(object):

    """ Metadata parser for NeXus files
    """

    def __init__(self, root):
        """ constructor

        :param root: nexus root node
        :type root: :class:`filewriter.FTGroup`
        """

        #: (:obj:`list` <:obj:`dict` <:obj:`str`, `any`> >) \
        #  description list of found nodes
        self.description = []
        #: (:obj:`str`) group postfix
        self.group_postfix = ""
        #: (:obj:`bool`) store NXentries as scientificMetadata
        self.scientific = False
        #: (:obj:`bool`) add empty units
        self.emptyunits = False

        #: (:obj:`list` <:obj:`str` >) \
        #    nexus field attribute show names
        self.attrs = None
        #: (:obj:`list` <:obj:`str` >) \
        #    nexus field attribute hidden names
        self.hiddenattrs = [
            "nexdatas_source",
            "nexdatas_strategy"
        ]
        #: (:obj:`list` <:obj:`str` >) \
        #    nexus entry classes to be shown
        self.entryclasses = [
            "NXentry"
        ]
        #: (:obj:`list` <:obj:`str` >) \
        #    nexus entry names to be shown
        self.entrynames = [
        ]
        #: (:obj:`dict` <:obj:`str`, [:obj:`str, `any`] > >) \
        #  attribute description
        self.attrdesc = {
            "nexus_type": ["type", str],
            "units": ["units", str],
            "depends_on": ["depends_on", str],
            "trans_type": ["transformation_type", str],
            "trans_vector": ["vector", str],
            "trans_offset": ["offset", str],
            "source_name": ["nexdatas_source", getdsname],
            "source_type": ["nexdatas_source", getdstype],
            "source": ["nexdatas_source", getdssource],
            "strategy": ["nexdatas_strategy", str],
        }
        #: (:obj:`dict` <:obj:`str`, [:obj:`str, `any`] > >) \
        #  metadata attribute description
        self.mattrdesc = {
            "source_name": ["nexdatas_source", getdsname],
            "source_type": ["nexdatas_source", getdstype],
            "source": ["nexdatas_source", getdssource],
            "strategy": ["nexdatas_strategy", str],
            "unit": ["units", str],
        }
        #: (:obj:`list`< :obj:`str`>)  field names which value should be stored
        self.valuestostore = ["depends_on"]
        self.__root = root
        #: (:obj:`list`< :obj:`str`>)  filters for `full_path` names
        self.filters = []
        # (:obj:`bool`) oned value flag
        self.oned = False
        # (:obj:`int`) maximal 1d record size
        self.maxonedsize = -1

    @classmethod
    def getpath(cls, path):
        """ converts full_path with NX_classes into nexus_path

        :param path: nexus full_path
        :type path: :obj:`str`
        """
        spath = path.split("/")
        return "/".join(
            [(dr if ":" not in dr else dr.split(":")[0])
             for dr in spath])

    def __addnode(self, node, tgpath):
        """adds the node into the description list

        :param node: nexus node
        :type node: :class:`filewriter.FTField` or \
                    :class:`filewriter.FTLink` or \
                    :class:`filewriter.FTAttribute` or \
                    :class:`filewriter.FTGroup`
        :param tgpath: target path of the link target or `None`
        :type tgpath: :obj:`str`
        """
        desc = {}
        path = filewriter.first(node.path)
        desc["full_path"] = str(path)
        desc["nexus_path"] = str(self.getpath(path))
        if hasattr(node, "dtype"):
            desc["dtype"] = str(node.dtype)
        if hasattr(node, "shape"):
            desc["shape"] = [int(n) for n in (node.shape or [])]
        if hasattr(node, "attributes"):
            attrs = node.attributes
            anames = [at.name for at in attrs]
            for key, vl in self.attrdesc.items():
                if vl[0] in anames:
                    desc[key] = vl[1](filewriter.first(attrs[vl[0]].read()))
        if node.name in self.valuestostore and node.is_valid:
            try:
                vl = node.read()
                cont = True
                while cont:
                    try:
                        if isinstance(vl, np.ndarray) and \
                           vl.shape == ():
                            vl = vl.item()
                            cont = False
                        elif not isinstance(vl, str) and \
                                (hasattr(vl, "__len__") and len(vl) == 1):
                            vl = vl[0]
                        else:
                            cont = False
                    except Exception:
                        cont = False
                desc["value"] = vl
            except Exception:
                pass
        self.description.append(desc)
        if tgpath:
            fname = self.__root.parent.name
            if "%s:/%s" % (fname, desc["nexus_path"]) != tgpath:
                ldesc = dict(desc)
                if tgpath.startswith(fname):
                    tgpath = tgpath[len(fname) + 2:]
                ldesc["nexus_path"] = "\\-> %s" % tgpath
                self.description.append(ldesc)

    def __parsenode(self, node, tgpath=None):
        """parses the node and add it into the description list

        :param node: nexus node
        :type node: :class:`filewriter.FTField` or \
                    :class:`filewriter.FTLink` or \
                    :class:`filewriter.FTAttribute` or \
                    :class:`filewriter.FTGroup`
        :param tgpath: target path of the link target or `None`
        :type tgpath: :obj:`str`
        """
        self.__addnode(node, tgpath)
        names = []
        if isinstance(node, filewriter.FTGroup):
            names = [
                (ch.name,
                 str(ch.target_path) if hasattr(ch, "target_path") else None)
                for ch in filewriter.get_links(node)]
        for nm in names:
            try:
                ch = node.open(nm[0])
                self.__parsenode(ch, nm[1])
#            except Exception:
#                pass
            finally:
                pass

    def __parsemetaentry(self, node, lst):
        """parses the node and add it into the description list

        :param node: nexus node
        :type node: :class:`filewriter.FTField` or \
                    :class:`filewriter.FTLink` or \
                    :class:`filewriter.FTAttribute` or \
                    :class:`filewriter.FTGroup`
        :param lst: metadata list
        :type lst: :obj:`dict` <:obj:`str`, `any`>
        """
        dct = {}
        name = self.__addmeta(node, dct, self.scientific)
        names = []
        if isinstance(node, filewriter.FTGroup):
            names = [
                (ch.name,
                 str(ch.target_path) if hasattr(ch, "target_path") else None)
                for ch in filewriter.get_links(node)]
        for nm in names:
            try:
                if name in dct.keys():
                    gr = dct[name]
                    if not isinstance(gr, dict):
                        nm = name + "_"
                        while nm in dct.keys():
                            nm = nm + "_"
                        dct[nm] = gr
                        gr = dct[name] = {}
                else:
                    gr = dct[name] = {}
                ch = node.open(nm[0])
                self.__parsemeta(ch, gr)
#            except Exception:
#                pass
            finally:
                pass
        lst.append(dct)

    def __parsemeta(self, node, dct):
        """parses the node and add it into the description list

        :param node: nexus node
        :type node: :class:`filewriter.FTField` or \
                    :class:`filewriter.FTLink` or \
                    :class:`filewriter.FTAttribute` or \
                    :class:`filewriter.FTGroup`
        :param dct: metadata dictionary
        :type dct: :obj:`dict` <:obj:`str`, `any`>
        """
        self.__addmeta(node, dct)
        names = []
        if isinstance(node, filewriter.FTGroup):
            names = [
                (ch.name,
                 str(ch.target_path) if hasattr(ch, "target_path") else None)
                for ch in filewriter.get_links(node)]
        for nm in names:
            try:
                name = node.name + self.group_postfix
                if name in dct.keys():
                    gr = dct[name]
                    if not isinstance(gr, dict):
                        nm = name + "_"
                        while nm in dct.keys():
                            nm = nm + "_"
                        dct[nm] = gr
                        gr = dct[name] = {}
                else:
                    gr = dct[name] = {}
                ch = node.open(nm[0])
                self.__parsemeta(ch, gr)
#            except Exception:
#                pass
            finally:
                pass

    def __addmeta(self, node, dct, scientific=False):
        """adds the node into the description list

        :param node: nexus node
        :type node: :class:`filewriter.FTField` or \
                    :class:`filewriter.FTLink` or \
                    :class:`filewriter.FTAttribute` or \
                    :class:`filewriter.FTGroup`
        :param dct: metadata dictionary
        :type dct: :obj:`dict` <:obj:`str`, `any`>
        :param scientific: scientific flag
        :type scientific: :obj:`bool`
        """
        desc = {}
        # path = filewriter.first(node.path)
        # desc["full_path"] = str(path)
        # desc["nexus_path"] = str(self.getpath(path))
        if isinstance(node, filewriter.FTGroup):
            if scientific:
                smname = "scientificMetadata"
                counter = 1
                while smname in dct.keys():
                    counter += 1
                    smname = "scientificMetadata_%s" % counter

                nd = dct[smname] = {"name": node.name}
            else:
                smname = node.name + self.group_postfix

                if smname in dct.keys():
                    nd = dct[smname]
                    if not isinstance(nd, dict):
                        nm = smname + "_"
                        while nm in dct.keys():
                            nm = nm + "_"
                        dct[nm] = nd
                        nd = dct[smname] = {}
                else:
                    nd = dct[smname] = {}
        else:
            smname = node.name
            if smname in dct.keys():
                nd = dct[smname]
                if not isinstance(nd, dict):
                    nm = smname + "_"
                    while nm in dct.keys():
                        nm = nm + "_"
                    dct[nm] = nd
                    nd = dct[smname] = {}
            else:
                nd = dct[smname] = {}
        if hasattr(node, "dtype"):
            desc["dtype"] = str(node.dtype)
        if hasattr(node, "shape"):
            desc["shape"] = [int(n) for n in (node.shape or [])]
        if hasattr(node, "attributes"):
            attrs = node.attributes
            anames = [at.name for at in attrs]
            for key, vl in self.mattrdesc.items():
                if vl[0] in anames and \
                   (self.attrs is None or key in self.attrs) and \
                   (self.hiddenattrs is None or key not in self.hiddenattrs):
                    nd[key] = vl[1](filewriter.first(attrs[vl[0]].read()))

            if self.attrs is not None:
                for at in self.attrs:
                    if at in anames:
                        if at in self.attrs and \
                           at not in self.mattrdesc.keys() and \
                           (self.hiddenattrs is None or
                                at not in self.hiddenattrs):
                            nd[at] = filewriter.first(attrs[at].read())
            else:
                for at in anames:
                    if at not in self.mattrdesc.keys() and \
                       (self.hiddenattrs is None or
                            at not in self.hiddenattrs):
                        nd[at] = filewriter.first(attrs[at].read())
            if self.scientific and "NX_class" in nd.keys() and \
               nd["NX_class"] == "NXentry":
                nd.pop("NX_class")
        if not isinstance(node, filewriter.FTGroup):
            if (node.name in self.valuestostore and node.is_valid) \
               or "shape" not in desc \
               or desc["shape"] in [None, [1], []] \
               or ((self.oned)
                   and len(desc["shape"]) == 1):
                if hasattr(node, "read"):
                    try:
                        vl = node.read()
                        cont = True
                        while cont:
                            try:
                                if isinstance(vl, np.ndarray) and \
                                   vl.shape == ():
                                    vl = vl.item()
                                    cont = False
                                elif not isinstance(vl, str) and \
                                        (hasattr(vl, "__len__") and
                                         len(vl) == 1):
                                    vl = vl[0]
                                else:
                                    cont = False
                            except Exception:
                                cont = False
                        if self.maxonedsize >= 0 and len(desc["shape"]) == 1 \
                           and hasattr(vl, "__len__") and \
                           len(vl) > self.maxonedsize:
                            try:
                                nd["value"] = [min(vl), max(vl)]
                            except Exception:
                                nd["value"] = [vl[0], vl[-1]]
                        else:
                            nd["value"] = vl
                        if self.emptyunits and "unit" not in nd.keys():
                            nd["unit"] = ""
                    except Exception:
                        pass
            if "shape" in desc and desc["shape"] not in [None, [1], []]:
                if "shape" in nd.keys():
                    shp = nd["shape"]
                    nm = "shape" + "_"
                    while nm in nd.keys():
                        nm = nm + "_"
                    nd[nm] = shp
                nd["shape"] = desc["shape"]
        return smname

    def __filter(self):
        """filters description list

        """
        res = []
        if self.filters:
            for elem in self.description:
                fpath = elem['full_path']
                found = False
                for df in self.filters:
                    found = fnmatch.filter([fpath], df)
                    if found:
                        break
                if found:
                    res.append(elem)
            self.description[:] = res

    def parse(self):
        """parses the file and creates the filtered description list

        """
        self.__parsenode(self.__root)
        self.__filter()

    def parseMeta(self):
        """parses the file and creates the filtered description list

        """
        for entry in self.__root:
            nm = entry.name
            at = None
            try:
                if "NX_class" in entry.attributes.names():
                    at = entry.attributes["NX_class"]
            except Exception:
                pass
            if len(self.entryclasses) == 0 or \
               at and (filewriter.first(at.read()) in self.entryclasses):
                if len(self.entrynames) == 0 or \
                   (nm and nm in self.entrynames):
                    self.__parsemetaentry(entry, self.description)


class FIOFileParser(object):

    """ Metadata parser for FIO files
    """

    def __init__(self, root):
        """ constructor

        :param root: fio file content
        :type root: :obj:`str`
        """

        #: (:obj:`list` <:obj:`dict` <:obj:`str`, `any`> >) \
        #  description list of found nodes
        self.description = []
        #: (:obj:`str`) group postfix
        self.group_postfix = ""
        #: (:obj:`dict` <:obj:`str`, `any`>)  metadata dictionary
        self.__dctmetadata = {}
        #: (:obj:`dict` <:obj:`str`, `any`>)  columns dictionary
        self.columns = {}

        # (:obj:`str`) text content of the file
        self.__root = root
        # (:obj:`bool`) oned value flag
        self.oned = False
        # (:obj:`int`) maximal 1d record size
        self.maxonedsize = -1

    def _appendComments(self, lines, meta):
        """append comments

        :param lines: comment fio lines
        :type lines: :obj:`list` <:obj:`str`>
        :param meta: metadata dictionary
        :type meta: :obj:`dict` <:obj:`str`, `any`>
        """
        comments = {}
        counter = 0
        for line in lines:
            if not line.startswith("!"):
                counter += 1
                comments["line_%s" % counter] = line
                if counter == 1:
                    meta["ScanCommand"] = line
            if "Acquisition started at " in line:
                sline = line.split("Acquisition started at ")
                if sline and sline[-1].strip():
                    meta["start_time"] = {
                        "value": isoDate(sline[-1].strip()),
                        "unit": ""
                    }
            elif "Acquisition ended at " in line:
                sline = line.split("Acquisition ended at ")
                if sline and sline[-1].strip():
                    meta["end_time"] = {
                        "value": isoDate(sline[-1].strip()),
                        "unit": ""
                    }
        if comments:
            meta["comments"] = comments

    def _appendParameters(self, lines, meta):
        """append comments

        :param lines: parameter fio lines
        :type lines: :obj:`list` <:obj:`str`>
        :param meta: metadata dictionary
        :type meta: :obj:`dict` <:obj:`str`, `any`>
        """
        params = {}
        for line in lines:
            if not line.startswith("!") and "=" in line:
                sline = line.split("=")
                if len(sline) > 1 and sline[0].strip() and \
                   sline[1].strip():
                    if '@' not in sline[0]:
                        try:
                            params[sline[0].strip().replace(" ", "_")] = \
                                eval(sline[1].strip())
                        except Exception:
                            params[sline[0].strip().replace(" ", "_")] = \
                                str(sline[1].strip())
        for line in lines:
            if not line.startswith("!") and "=" in line:
                sline = line.split("=")
                if len(sline) > 1 and sline[0].strip() and \
                   sline[1].strip():
                    if '@' in sline[0]:
                        field, attr = sline[0].strip().replace(" ", "_"). \
                            split("@")[:2]
                        try:
                            avl = eval(sline[1].strip())
                        except Exception:
                            avl = str(sline[1].strip())

                        if field in params:
                            if not isinstance(params[field], dict):
                                params[field] = {"value": params[field]}
                            if attr in ["unit", "units"]:
                                params[field]["unit"] = avl
                            else:
                                params[field][attr] = avl

        if params:
            meta["parameters"] = params

    def _appendData(self, lines, meta):
        """append comments

        :param lines: data fio lines
        :type lines: :obj:`list` <:obj:`str`>
        :param meta: metadata dictionary
        :type meta: :obj:`dict` <:obj:`str`, `any`>
        """
        self.columns = {}
        data = {}
        for line in lines:
            if line.startswith("Col"):
                sline = line.split(" ")
                name = None
                if len(sline) > 2:
                    try:
                        if sline[1].strip():
                            cid = int(sline[1].strip())
                            if sline[2].strip():
                                name = str(sline[2].strip())
                                self.columns[cid - 1] = [name, []]
                    except Exception:
                        pass
            elif not line.startswith("!"):
                sline = [word.strip() for word in line.split(" ")
                         if word.strip()]
                for wid, word in enumerate(sline):
                    if wid in self.columns.keys():
                        try:
                            self.columns[wid][1].append(float(word))
                        except Exception:
                            self.columns[wid][1].append(str(word))
        for wid, nmvl in self.columns.items():
            if self.maxonedsize >= 0 \
               and hasattr(nmvl[1], "__len__") and \
               len(nmvl[1]) > self.maxonedsize:
                data[nmvl[0]] = [nmvl[1][0], nmvl[1][-1]]
            else:
                data[nmvl[0]] = nmvl[1]
        if data:
            meta["data"] = data

    def parseMeta(self):
        """parses the file and creates the filtered description list

        """
        smname = "scientificMetadata"
        dct = {}
        self.description = [dct]
        nd = dct[smname] = {}
        if self.__root and isinstance(self.__root, str):
            lines = [line.strip() for line in self.__root.split("\n")]
            dcpmap = {"%d": [], "%c": [], "%p": []}
            last = None
            for line in lines:
                if line in dcpmap.keys():
                    last = line
                elif last and not line.startswith("!"):
                    dcpmap[last].append(line)
                elif line.startswith("!"):
                    dcpmap["%c"].append(line)

            if dcpmap["%c"]:
                self._appendComments(dcpmap["%c"], nd)
            if dcpmap["%p"]:
                self._appendParameters(dcpmap["%p"], nd)
            if dcpmap["%d"] and (self.oned):
                self._appendData(dcpmap["%d"], nd)



#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2017 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

""" Provides redis utils """


REDIS = True
try:
    from redis_om import HashModel, Field
    from typing import Optional
    from blissdata.redis_engine.identities import _UninitializedRedis
    from blissdata.redis_engine.store import DataStore
except Exception as e:
    print("Redis or blissdata cannot be imported: %s" % str(e))
    REDIS = False


if REDIS:

    class DESYIdentityModel(HashModel):
        """Institute specific information used to link scans
           in Redis to external services.
        """

        class Meta:
            global_key_prefix = "esrf"
            model_key_prefix = "id"
            database = _UninitializedRedis()

        name: str = Field(index=True)
        number: int = Field(index=True)
        data_policy: str = Field(index=True)

        # DESY data policy
        beamline: Optional[str] = Field(index=True)
        session: Optional[str] = Field(index=True)
        proposal: Optional[str] = Field(index=True)
        collection: Optional[str] = Field(index=True)
        dataset: Optional[str] = Field(index=True)

        # Without data policy
        path: Optional[str] = Field(index=True)

else:
    DESYIdentityModel = None


def getDataStore(redisURL):

    datastore = None
    try:
        datastore = DataStore(redisURL, init_db=True,
                              identity_model_cls=DESYIdentityModel)
    except Exception:
        print("Redis DataStore already initialized")
        try:
            datastore = DataStore(redisURL,
                                  identity_model_cls=DESYIdentityModel)
        except Exception as e:
            print("Redis DataStore cannot be instantiated: %s" % str(e))

    return datastore



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
# \package ndtstools tools for ndts
# \file testXMLCreator.py
# test of XML file creator

from nxstools.nxsxml import (
    XMLFile, NGroup, NDSource, NField, NDimensions, NDeviceGroup)


# the main function
def main():
    df = XMLFile("MNI.xml")

    en = NGroup(df, "entry2", "NXentry")

    ins = NGroup(en, "instrument", "NXinstrument")
#    NXsource
    src = NGroup(ins, "source", "NXsource")
    f = NField(src, "distance", "NX_FLOAT")
    f.setUnits("m")
    f.setText("100.")
    f = NField(src, "db_devices", "NX_CHAR")
    d = NDimensions(f, "2")
    d.dim("1", "151")
    d.dim("2", "2")
    f.setStrategy("STEP")
    sr = NDSource(f)
    sr.initDBase(
        "db_devices", "MYSQL",
        "SELECT name, pid FROM device limit 151", "tango", "IMAGE",
        host="haso228k.desy.de")
    f = NField(src, "type", "NX_CHAR")
    f.setText("Synchrotron X-ray Source")
    f = NField(src, "name", "NX_CHAR")
    f.setText("PETRA-III")
    f.addAttr("short_name", "NX_CHAR", "P3")
    f = NField(src, "probe", "NX_CHAR")
    f.setText("x-ray")
    f = NField(src, "power", "NX_FLOAT")
    f.setUnits("W")
    f.setText("1")
#    sr = NDSource(f)
#    sr.initTango("p09/motor/exp.01","p09/motor/exp.01", "attribute",
#                 "power", host="haso228k.desy.de", port="10000")
    f = NField(src, "emittance_x", "NX_FLOAT")
    f.setUnits("nm rad")
    f.setText("0.2")
    f.setStrategy("STEP")
    sr = NDSource(f)
    sr.initClient("emittance_x", "emittance_x")
    f = NField(src, "emittance_y", "NX_FLOAT")
    f.setUnits("nm rad")
    f.setText("0.2")
#    sr = NDSource(f)
#    f.setStrategy("STEP")
#    sr.initSardana("door1","door1", "emittance_y",
#             host="haso228k.desy.de", port="10000");
    f = NField(src, "sigma_x", "NX_FLOAT")
    f.setUnits("nm")
    f.setText("0.1")
    f = NField(src, "sigma_y", "NX_FLOAT")
    f.setUnits("nm")
    f.setText("0.1")
    f = NField(src, "flux", "NX_FLOAT")
    f.setUnits("s-1 cm-2")
    f.setText("0.1")
    f = NField(src, "energy", "NX_FLOAT")
    f.setUnits("GeV")
    f.setText("0.1")
    f = NField(src, "current", "NX_FLOAT")
    f.setUnits("A")
    f.setText("10")
    f = NField(src, "voltage", "NX_FLOAT")
    f.setUnits("V")
    f.setText("10")
    f = NField(src, "period", "NX_FLOAT")
    f.setUnits("microseconds")
    f.setText("1")
    f = NField(src, "target_material", "NX_CHAR")
    f.setText("C")
# any source/facility related messages/events
# that occurred during the experiment
    # g =
    NGroup(src, "notes", "NXnote")
# For storage rings, description of the bunch
# pattern. This is useful to describe irregular
# bunch patterns.
# See table: NXsource: NXsource/bunch_pattern:NXdata
    # g =
    NGroup(src, "bunch_pattern", "NXdata")
    f = NField(src, "number_of_bunches", "NX_INT")
    f.setText("1")
    f = NField(src, "bunch_length", "NX_FLOAT")
    f.setUnits("s")
    f.setText("1")
    f = NField(src, "bunch_distantce", "NX_FLOAT")
    f.setUnits("s")
    f.setText("1")
    f = NField(src, "bunch_width", "NX_FLOAT")
    f.setUnits("s")
    f.setText("2")
    # source pulse shape
    # g =
    NGroup(src, "pulse_shape", "NXdata")
    f = NField(src, "mode", "NX_CHAR")
    f.setText("Single Bunch")
    f = NField(src, "top_up", "NX_BOOLEAN")
    f.setText("1")

    src.setText("My source")
    #    mot1 = NDeviceGroup(ins, "p09/motor/exp.01", "motor1",
    #                  "NXpositioner", commands=False)
    mot1 = NDeviceGroup(ins, "p09/motor/exp.01", "motor1", "NXpositioner",
                        commands=False,
                        blackAttrs=["PositionEncoder", "PositionEncoderRaw"])
    mot2 = NDeviceGroup(ins, "p09/motor/exp.02", "motor2", "NXpositioner",
                        commands=False,
                        blackAttrs=["PositionEncoder", "PositionEncoderRaw"])
    # mca =
    NDeviceGroup(ins, "p09/mca/exp.02", "mca2", "NXdetector")
    # cnt =
    NDeviceGroup(ins, "p09/counter/exp.02", "counter", "NXmonitor")
    # dac =
    NDeviceGroup(ins, "p09/dac/exp.02", "dac", "NXsensor")
    # adc =
    NDeviceGroup(ins, "p09/adc/exp.02", "adc", "NXsensor")
    # vfc =
    NDeviceGroup(ins, "p09/vfc/exp.02", "vfcadc", "NXsensor")
    # dgg2 =
    NDeviceGroup(ins, "p09/dgg2/exp.01", "dgg2", "NXmonitor")
    tst = NGroup(en, "tst", "NXinstrument")
    mot1.setText("My motor1")
    mot2.setText("My motor1")

    NDeviceGroup(tst, "p09/tst/exp.01", "Tst", "NXmonitor")

    df.dump()


if __name__ == "__main__":
    main()
